/* ***** BEGIN LICENSE BLOCK *****
 * Licensed under Version: MPL 1.1/GPL 2.0/LGPL 2.1
 * Full Terms at http://mozile.mozdev.org/license.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Playsophy code (www.playsophy.com).
 *
 * The Initial Developer of the Original Code is Playsophy
 * Portions created by the Initial Developer are Copyright (C) 2002-2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Richard Prescott, revised on 2004-04-18, Added some namespace support.
 *
 * ***** END LICENSE BLOCK ***** */

/**
@fileoverview 
eDOM.js editor or extended DOM

<p>The basic w3c DOM is not honed for interactive editing. eDOM extends the standard DOM
to make it easy to build CSS-enabled XML and XHTML editing applications that work
within a browser. Initially it should support editing of XHTML in XHTML and XML pages 
within any W3c DOM supporting browser. Geiko-based browser like Mozilla are the initial 
target.

<p>When complete, eDOM will be a DOM module that facilities a wide variety of domain specific
and general purpose editors. It itself is not an editor: issues like Selection models
and toolbars are beyond its scope. It represents the routines needed in editors built
over the w3c DOM. First releases tackle text/object manipulation (editing and styling); 
later releases will address layout editing.
 
<p>The initial target is basic XHTML editing. This requires some new methods for w3c DOM
objects and three new Objects:
<ul>
<li><b>CSSRange</b>: range content may be manipulated in various ways; CSS is one
way to restrict how it should be processed.
<li><b>InsertionPoint</b>: allow you to iterate, examine and manipulate the visible characters 
of a document
<li><b>CSSLine</b>: captures the set of elements that are rendered as a line on a screen
</ul>

<p>POST05:
<ul>
<li>all XHTML stuff out of here and into eDOMXHTML
<li> - redo tighten Range methods to use lines and insertion points properly
<li> - new restoration methods for insertion point and range (TextPointer is wasteful)
<li> - ability to inline style text based on an inline element like EM or STRONG as opposed
to explicitly applying inline styles.
<li> - dom level 3 worked in: ie/ validate line or insertion methods based on what schema
allows. Do first for XHTML DL. May make separate ie/ pure CSS driven methods of eDOM
allow anything CSS permits but can ask validation method on CSSLine etc as they
should have methods that tie into domlevel3.
</ul>

<p>Max : Version 0.6.0 is my development version for solid XHTML editing.
Not guaranteed to work correctly or support any other version...

@version 0.6.0
*/

/*********************************** eDOM version *********************************/

var eDOM_VERSION = "0.6.0";

/**
Is there support for the eDOM either as a whole or for a particular version.

<p>Use: if(document.supportsEDOM)<br />
or if(document.supportsEDOM("0.5"))

@argument {String} version Optional.
@return true if eDOM or the specified <code>version</code> is supported.
@type boolean
 */
Document.prototype.supportsEDOM = function(version)
{
	if ( !arguments[0] )
		return true;

	//There is a possibility that an eDOM version can fully support other earlier version numbers.
	//If so, this needs to be added here.
	if( version == eDOM_VERSION ) {
		return true;
	} else {
		return false;
	}
}

/*********************************** Node/Element/Text/CSS ***************************
 * POST05: Range rework will remove last XHTML references and a number of the following
 * methods.
 * 
 *************************************************************************************/

/**
eDOM : Gets a node's offset within its parent.

@return offset
@type integer
*/
Node.prototype.__defineGetter__(
	'offset',
	function()
	{
		var parentNode = this.parentNode;
		var nextChild = parentNode.firstChild;
		var offset = 0;
		while(nextChild)
		{
			if(nextChild == this)
				break;
			offset++;
			nextChild=parentNode.childNodes[offset];
		}
		return offset;
	}
);

/**
eDOM : Gets the parent Element for this node.

<p>If this node is a Text Node, or an element that doesn't allow content, return the parentNode.
Otherwise return this node.

@return parentElement
@type Node
*/
Node.prototype.__defineGetter__(
	'parentElement',
	function()
	{
		if((this.nodeType == Node.TEXT_NODE) || (this.contentType == Element.EMPTY_CONTENTTYPE))
			return this.parentNode;
		return this;
	}
);

/**
eDOM : Gets the top inline ancestor node of this node.
<p>i.e the first child of the first ancestor node that does not have
<code>CSS display == inline</code>. Could be this node.

@return topInlineAncestor
@type Node
*/
Node.prototype.__defineGetter__(
	'topInlineAncestor',
	function()
	{
		var nodeToCheck = this.parentNode;
		if(nodeToCheck.computedStyleValue("display") == "inline")
			return nodeToCheck.topInlineAncestor;
		return this;
	}
);

/**
eDOM : Gets the first not inline <code>parentElement</code>.
i.e. The first parent element that doesn't have the computed style <code>display == inline</code>.

@return nilParent
@type Node
*/
Node.prototype.__defineGetter__(
	"__nilParentElement",
	function()
	{
		var nilParent = this.parentElement;
		while( (nilParent.computedStyleValue("display") == "inline") && (nilParent != document.documentElement) )
			nilParent = nilParent.parentNode;
		return nilParent;
	}
);

/**
eDOM : Is this node a descendent of ancestor?
<p>Returns true if this node is a descendent of ancestor, otherwise false.

@param {Node} ancestor
@return descendent
@type boolean
*/
Node.prototype.descendent = function(ancestor)
{
	var nodeParent = this.parentNode;
	while(nodeParent)	
	{
		if(nodeParent == ancestor)
			return true;
		nodeParent = nodeParent.parentNode;
	}
	return false;
}

/**
eDOM : inserts newParent node before this node.

<p>Only used in eDOMXHTML.js for list elements to add new list parent. e.g. ul, ol. 

<p>POST05:
If you use this the node gets lost from any range it is within.
i.e. DON'T USE THIS IF YOU NEED TO BE ABLE TO RESTORE A MARKED TEXT RANGE.
This needs to be reworked as a method (of the parent node) so that the
range is not affected. 

@param {Node} newParent node to insert.
@return inserted parent node.
@type Node
*/
Node.prototype.insertParent = function(newParent)
{
	try {
	var currentParent = this.parentNode;
	newParent = currentParent.insertBefore(newParent, this);
	newParent.appendChild(this);
	return newParent;
	}catch(e){
		alert("eDOM error, Node.insertParent "+e);
		return newParent;
	}
}

/**
eDOM : Gets the computed style value.

<p>This returns the default view's computed style for a property with the following qualifications:
<ul>
<li>for text nodes, for "display" it returns "inline"; for other properties it returns their parent's value.
<li>for XHTML &lt;br/&gt;, "display" is set to "block" irrespective of what the default says it is.
 &lt;br/&gt; always behaves as a block.
 I don't like this hack but Mozilla (and perhaps other browsers) treat &lt;br/&gt;
 as inline when it is obviously a block level element.
</ul>

<p>Note: ideally this should be .computedStyle and then use getPropertyValue etc.
However, the DOM doesn't let me create CSS Style definitions explicitly.

<p>MAX: changed br test from nodeName to localName.
When in a Heading element br returns display==inline.

@param {String} propertyName CSS property
@return Default View computed style value
@type String
*/ 
Node.prototype.computedStyleValue = function(propertyName) {
	var nodeToCheck = this;
	if(this.nodeType == Node.TEXT_NODE) {
		if(propertyName == "display")
			return "inline";
		nodeToCheck = this.parentNode;
	}
	if(this.nodeType == Node.ELEMENT_NODE) {
		var cs = document.defaultView.getComputedStyle(nodeToCheck, null);
		var csv = cs.getPropertyValue(propertyName);
//	if( propertyName=="display" && (this.localName.toLowerCase()=="br") )
//		csv="block";
		return csv
	} else {
		return null;
	}
}

/**
eDOM : Gets the previous editable sibling (Identical to editablePreviousSibling).
 
<p>Returns the first previous sibling that is not an empty text node, or null if there
isn't an editable previous sibling.

<p>POST04: need to change to use IPNodeFilter (get rid of nonEmpty check everywhere)
- will need to account for skip here! Careful - need to see if rework works in normalize etc

@returns editablePreviousSibling
@type Node
*/
Node.prototype.__defineGetter__(
	"__editablePreviousSibling",
	function()
	{
		var previousSibling = this.previousSibling;
		while(previousSibling)
		{
			if(__NodeFilter.nonEmptyText(previousSibling) == NodeFilter.FILTER_ACCEPT)
				return previousSibling;
			previousSibling = previousSibling.previousSibling;
		}
		return null;
	}
);

/**
eDOM : Gets the next editable sibling (Identical to editableNextSibling).
<p>An editable sibling is one with at least one non-empty text node as a descendent.
<p>Returns non empty node after this one or null if there isn't one. 
<p>POST04: need to change to use IPNodeFilter

@returns editableNextSibling
@type Node
*/
Node.prototype.__defineGetter__(
	"__editableNextSibling",
	function()
	{
		var nextSibling = this.nextSibling;
		while(nextSibling)
		{
			if(__NodeFilter.nonEmptyText(nextSibling) == NodeFilter.FILTER_ACCEPT)
				return nextSibling;
			nextSibling = nextSibling.nextSibling;
		}
		return null;
	}
);

/**
eDOM : Gets the previous editable sibling.

<p>Returns the first previous sibling that is not an empty text node, or null if there isn't one.
 
<p>POST04: need to change to use IPNodeFilter (get rid of nonEmpty check everywhere) - will need to account for
skip here! Careful - need to see if rework works in normalize etc

<p>Could use contentType for Element but need to dig deeper if contentType is "_ELEMENT". If ANY then skip element.

@returns editablePreviousSibling
@type Node
*/
Node.prototype.__defineGetter__(
	"editablePreviousSibling",
	function()
	{
		var previousSibling = this.previousSibling;
		while(previousSibling)
		{
			if(__NodeFilter.nonEmptyText(previousSibling) == NodeFilter.FILTER_ACCEPT)
				return previousSibling;
			previousSibling = previousSibling.previousSibling;
		}
		return null;
	}
);

/**
eDOM : Gets the next editable sibling.
An editable sibling is one with at least one non-empty text node as a decendent.
<p>Returns non empty node after this one or null if there isn't one
<p>POST04: need to change to use IPNodeFilter

@returns editableNextSibling
@type Node
*/
Node.prototype.__defineGetter__(
	"editableNextSibling",
	function()
	{
		var nextSibling = this.nextSibling;
		while(nextSibling)
		{
			if(__NodeFilter.nonEmptyText(nextSibling) == NodeFilter.FILTER_ACCEPT)
				return nextSibling;
			nextSibling = nextSibling.nextSibling;
		}
		return null;
	}
);


Node.prototype.editableFirstChild =	function()
{
	if ( !this.hasChildNodes() )
		return null;
	var c = this.childNodes;
	var l = c.length;
	for (var i=0; i<l; i++) {
		if(__NodeFilter.nonEmptyText(c[i]) == NodeFilter.FILTER_ACCEPT)
			return c[i];
	}
	return null;	
}

/**** ELEMENT NODES ****/

/**
eDOM : Removes all the child nodes of this Element.
*/ 
Element.prototype.removeAllChildren = function() 
{
	var child = this.firstChild;        
	while (child) {
		var oldchild = child;
		child = child.nextSibling;
		this.removeChild(oldchild);        
	}
}

/**
eDOM : inserts <code>newChild</code> after<code>refChild</code>.

<p>Peer of <code>Node.insertBefore</code>.
 
@param {Node} newChild
@param {Node} refChild
@return newChild or DOMException
@type Node
 */
Node.prototype.insertAfter = function(newChild, refChild)
{
	if(this.hasChildNodes() && this.lastChild == refChild)
		return this.appendChild(newChild);
	else
		return this.insertBefore(newChild, refChild.nextSibling);
}

/**
eDOM : Splits the element in two with the children before offset staying with this element
and all the following children going to a new sibling.

<p>Split won't happen if offset points before the first, or after the last,
editable child of the element.

<p>Offset is the zero based index into the list of child nodes.
The children BEFORE offset stay with this node.

@param {integer} offset
@return new element inserted after this element in the tree or null if no split happens
@type Node
*/
Element.prototype.split = function(offset)
{
	// If offset is too big and doesn't exist can't split.
	if(this.childNodes.length <= offset)
		return null;

	// If no previous editable sibling before offset then nothing to split.
	if(!this.childNodes[offset].editablePreviousSibling)
		return null;

	// If no editable sibling at or beyond offset do nothing. (Note: offset always > 0 because of previous test.)
	if(!this.childNodes[offset-1].editableNextSibling)
		return null;

	var newElement = this.cloneNode(false);
	this.parentNode.insertAfter(newElement, this);
	var child;
//	while( this.hasChildNodes() && (this.childNodes.length >= offset) )
// MAX : second test appears to be wrong. if >= then on the last iteration this.childNodes[offset] is null.
	while( this.hasChildNodes() && (this.childNodes.length > offset) )
	{
		child = this.childNodes[offset];
		newElement.appendChild(child);
	}
	return newElement;
}

/**
eDOM : XHTML specific : replace child with newChildName using XHTML namespace.

Surely this should be named replaceChildOnlyXHTML or something similar?
This should only be used if you really do want the XHTML namespace included.
i.e. don't use for text/html documents.
*/
Element.prototype.replaceChildOnly = function(child, newChildName)
{
	return this.replaceChildOnlyNS(child, eDOMXHTML.NS, newChildName);
}
/**
eDOM : replace child with newChildName, using newChildNamespace if specified.

@param {Node} child
@param {String} newChildNamespace - optional
@param {String} newChildName
@return newElement
@type Element
*/
Element.prototype.replaceChildOnlyNS = function(child, newChildNamespace, newChildName)
{
	var newElement = eDOMXHTML.createElementNS(newChildNamespace, newChildName);
	// copy attributes of this element
	for(var i=0; i<child.attributes.length; i++)
	{
		var childAttribute = child.attributes.item(i);
		var childAttributeCopy = childAttribute.cloneNode(true);
		newElement.setAttributeNode(childAttributeCopy);
	}
	var childContents = document.createRange();
	childContents.selectNodeContents(child);

	newElement.appendChild(childContents.extractContents());
	childContents.detach();
	this.replaceChild(newElement, child);
	return newElement;
}

Element.prototype.removeChildOnly = function(oldChild)
{
	if(oldChild.childNodes.length > 0)
	{
		var childContents = document.createRange();
		childContents.selectNodeContents(oldChild);
		this.insertBefore(childContents.extractContents(), oldChild);
	}

	return this.removeChild(oldChild); 
}

/**
eDOM: Create an insertion point for the first insertion point of this element.
Note: this will insert a token to allow selection if need be.

@param top
@return ip
@type InsertionPoint
*/
Element.prototype.firstInsertionPoint = function(top)
{
	var firstIP;

	if(this.contentType == Element.EMPTY_CONTENTTYPE) {
	// special case: first insertion point in an empty element is right after that element
	//MAX: BUG!!! BUT this.offset+1 doesn't exist if at end of line...
		firstIP = __createInsertionPoint(top, this.parentNode, (this.offset));
		firstIP.forwardOne();
		return firstIP;
	}
	//else not an empty element.

	var __ipni = new __IPNodeIterator(this);
	if(__ipni.currentNode) {
		if ( __ipni.currentNode.nodeType == Node.TEXT_NODE )
			firstIP = __createInsertionPoint(top, __ipni.currentNode, 0);
		else
			firstIP = __createInsertionPoint(top, __ipni.currentNode.parentNode, __ipni.currentNode.offset);
		return firstIP;
	}	
	//else we add a token.
	var textNode = eDOM.token();
	this.appendChild(textNode);
	var onlyip = __createInsertionPoint(top, textNode, 0);
	return onlyip;
}

/**
eDOM: Create an insertion point for the last insertion point of this element.
Note: this will insert a token to allow selection if need be.

@param top
@return ip
@type InsertionPoint
*/
Element.prototype.lastInsertionPoint = function(top)
{
	var __ipni = new __IPNodeIterator(this);
	__ipni.setToEnd();
	if(__ipni.currentNode)
	{
		var lastip = (__ipni.currentNode.nodeType == Node.TEXT_NODE) ? __createInsertionPoint(top, __ipni.currentNode, __ipni.currentNode.nodeValue.length) : __createInsertionPoint(top, __ipni.currentNode.parentNode, (__ipni.currentNode.offset+1));
		return lastip;
	}	

	var textNode = eDOM.token();
	this.appendChild(textNode);
	var onlyip = __createInsertionPoint(top, textNode, 0);
	return onlyip;
}

/*
 * POST04: 
 * - account for "class" as well as style (treat as generic enum attr where order
 * doesn't matter)
 * - account for "display": never match table columns
 * - xml: do full attribute matching
 * - xml: style doesn't come from a style method
 */
Element.prototype.match = function(node)
{
	if(!node)
		return false;

	if(this.nodeType != node.ELEMENT_NODE)
		return false;

	if(this.nodeName.toLowerCase() != node.nodeName.toLowerCase())
		return false;

	if(this.attributes.length != node.attributes.length)
		return false;

	for(var i=0; i<this.attributes.length; i++)
	{
		var thisAttribute = this.attributes.item(i);
		var nodeAttribute = node.attributes.getNamedItem(thisAttribute.nodeName);
		if(!nodeAttribute)
			return false;

		// do literal string compare on all but style attribute. This may not work with
		// XML which may have other attributes like style. Need more sophisticated attribute
		// match handling.

		// POST04: this may not work for XML. Seem to remember not having style explicitly.
		if(nodeAttribute.nodeName == "style")
		{
			if(!node.style.match(this.style))	
				return false;
		}
		// assume all but style need to be exactly the same!	
		else if(nodeAttribute.nodeValue != thisAttribute.nodeValue)
		{
			return false;
		}		
	}	

	return true;
}

/**
 * Set the value of style
 *
 * POST04:
 * - distinguish inline and block level styles ...
 * - may move most of this into "CSSLine.setStyle": list all styles, see if any integer styles for inline
 * - needs to use style property meta; now only works for integer value styles
 * [meta: name, displays, int or enum or ..., units (if int) etc
 * - text justify: CSSLine ... text-align (justify is default?)
 */
Element.prototype.setStyle = function(styleName, styleValue)
{
	var intStyleRegExp = /(\D?)(\d+)(.*)/i;
	var regResult = intStyleRegExp.exec(styleValue);

	// non integer style - just set the value for now
	if(!regResult)
	{
		this.style.setProperty(styleName, styleValue, "");
		return;
	}

	// integer style from here on ...
	var plusOrMinus = regResult[1];
	var intVal = parseInt(regResult[2]);
	var units = regResult[3];

	// style attribute value is a string (this.attributes.getNamedItem("style").value). This is far from ideal
	// for easy setting: basically reduces style manip to string matching. For now use, HTML style attribute
	// which gives a CSSDeclaration ... Not ideal as doesn't work for XML.

	// shouldn't this come from computedValue? POST04 ie/ factor in inherited offsets ...
	var currentStyleValue = parseInt(this.style.getPropertyValue(styleName));
	if(isNaN(currentStyleValue))
	{
		if(plusOrMinus != "-") // for now, don't allow for - values
			this.style.setProperty(styleName, intVal+units, "");
		return;
	}
	
	if(plusOrMinus == "+") // increment
		currentStyleValue += intVal;
	else if(plusOrMinus == "-") // decrement
	{
		// for now, allow no negatives!
		if(currentStyleValue <= intVal)
		{
			this.style.removeProperty(styleName);
			// if style is now empty, remove it!
			if(this.style.length == 0) // nix style if no other style setting!
				this.attributes.removeNamedItem("style");
			return;
		}

		currentStyleValue -= intVal;		
	}

	this.style.setProperty(styleName, currentStyleValue + units, "");
}

/**
 * Does a node have a particular name - case insensitive
 * 
 * POST05: make into accessor giving back lower case name every time
 */
Element.prototype.nodeNamed = function(nodeName)
{
	return(this.nodeName.toLowerCase() == nodeName.toLowerCase());
}

/*
 * POST04: need to account for synonyms and complex css values
 */
CSSStyleDeclaration.prototype.match = function(declToMatch)
{
	if(this.length != declToMatch.length)
		return false;

	for(var j=0; j<this.length; j++)
	{
		var aStyleName = this.item(j);
		var bStyleValue = declToMatch.getPropertyValue(aStyleName);
		var aStyleValue = this.getPropertyValue(aStyleName);
		if(aStyleValue != bStyleValue)
		{
			return false;
		}				
	}

	return true;
}

/**
Is value2 equal to, or a synonym for, value.
Override this function with an application specific version.
To do : make more generic, this is a simple starting point.

@param {String} name style name
@param {String} value style value
@param {String@ value2 alternate style value
@return true if styles are the same.
@type Boolean 
*/
EDOM.prototype.compareStyleValues = function( name, value, value2 ) {
	return ( value2 == value );
}

// ========================== InsertionPoint =================================== 

/**
Insertion Point
<p>An IP or Insertion Point node is:
<ul>
<li>a point within a visible text node
<li>on or right after an element node of EMPTY_CONTENTTYPE. The element can have inline or block
display
</ul>

<h3>Text</h3>

<p>Text rendering varies according to the handling of whitespace. All flavors of XML distinguish
between whitespace that displays and whitespace that makes page markup more presentable.
 
<p>As the XML standard says: "In editing XML documents, it is often convenient to use 
"white space" (spaces, tabs, and blank lines) to set apart the markup for greater 
readability. Such white space is typically not intended for inclusion in the delivered 
version of the document. On the other hand, "significant" white space that should be 
preserved in the delivered version is common, for example in poetry and source code."

<p>eDOM must traverse the rendered whitespace based on the actual whitespace in the "raw"
text.
<ul>
<li>HTML: handle whitespace = &lt;pre&gt;: http://www.w3.org/TR/html4/struct/text.html#h-9.3.4 
<li>XML: http://www.w3.org/TR/2000/REC-xml-20001006#sec-white-space
<li>XHTML: http://www.w3.org/TR/xhtml1/#uaconf
</ul>

<p>eDOM's whitespace handling is based on the CSS "white-space" property. InsertionPoint skips 
invisible whitespace.
 
<h3>EMPTY Elements</h3>

<p>Empty elements such as xhtml:img or xhtml:hr may display as blocks or within lines. In both cases,
you may wish to insert text or other empty elements before or after them - in other words, there are
insertion points immediately before or after these elements.

<h3>Tokens</h3>

<p>Geiko-based browsers don't display childless, content bearing elements. 

<p>eDOM supports the concept of an "InsertionPoint Token" (IPT) whose presence makes Geiko show
an element and for eDOM marks that insertion is allowed within an otherwise empty element. 
The first inserted item will replace, as opposed to adding to, the token. 

<p>Valid tokens are:
<ul>
<li>an XHTML <br/> (invisible element) in a line with nothing other than whitespace [TODO]
<li>a nbsp; on its own in a line
</ul>

<p>When laying out a page a user can choose to put nothing but tokens in various editable elements.

<h3>New lines, same line and at top</h3>

<p>InsertionPoint's forward and back methods return whether they keep the point within a line or
move it to a new line or remain in the same position because the point is now at its top boundary.

<p>"Top" is important: it is used to restrict a pointer within an area so that editing will only take
place within that area and that the area remains editable irrespective of what operations are performed
on it through eDOM.

<h4>POST05</h4>
<ul> 
<li>more robust: check that its ipnode is valid; if not then reset to first in top!
<li>consider rename CROSSED_BLOCK to CROSSED_LINE or "NEW_LINE"
<li>add replaceCharacter (insert mode)
<li>iptoken (expand to cover BR etc)
<li>consider getters for "endOfContainer", "startOfContainer" to cover whether an IP starts
or ends its inline or block level container
</ul>

<p>POST04: need to accept selection of inline isolated whitespace
*/
function InsertionPoint(top, seed, seedOffset, cw, csswsp)
{
	this.__top = top;
	this.__ipNode = seed;
	this.__ipOffset = seedOffset;
	this.__cw = cw;
	this.__cssWhitespace = csswsp;
	this.__markedOffset = -1; // no setting unless mark is made
}

/**
0.
@final
*/
InsertionPoint.SAME_LINE = 0;
/**
1.
@final
*/
InsertionPoint.CROSSED_BLOCK = 1;
/**
2.
@final
*/
InsertionPoint.AT_TOP = 2;


/**
Create Insertion Point.
 
<p>Only three types of valid seed:
<ol> 
<li>A text node, provided it is not whitespace only. (An editing token (nbsp;) is ok.)
<li>A parent and offset that points to an empty element node. The element node can be block or inline.
<li>A parent and offset that points immediately AFTER an empty element node. The element node can be block or inline.
<ol>

<p>Creation corrects the default IP for five different whitespace issues:
<ol>
<li>middle of collapsed midline whitespace
<li>within or before the whitespace of a start of line non whitespace text node that begins with whitespace
<li>within or after the whitespace of an end of line non whitespace text node that ends with whitespace
<li>a whitespace only text node that starts a line (skip after)
<li>a whitespace only text node that ends a line (skip before)
</ol>

<p>POST05:<br />
 - check for seed validity (move up from Range.firstInsertionPoint)
*/
Document.prototype.createInsertionPoint = documentCreateInsertionPoint;

function documentCreateInsertionPoint(top, seed, seedOffset)
{	
	// a/c for special case of a reference to a text node rather than a reference within it
	if((seed.nodeType == Node.ELEMENT_NODE) && (seed.childNodes.length > seedOffset) && (seed.childNodes[seedOffset].nodeType == Node.TEXT_NODE))
	{
		seed = seed.childNodes[seedOffset];
		seedOffset = 0;	
	}

	var ip = __createInsertionPoint(top, seed, seedOffset);

	// can't be within or before whitespace at the start of a line;
	// ...within whitespace within the line;
	// or within or after whitespace at the end of a line
	if((ip.__cssWhitespace != "pre") && ip.whitespace)
	{
		// go to start of whitespace sequence or token (same_line) or end of previous block (crossed_block) or stay where you are (at_top)
		var result = ip.backOne();

		// special case: whitespace only text node that COULD BE at the end of the line. Jump after previous valid ip
		if((result == InsertionPoint.SAME_LINE) && ip.whitespace && (ip.ipOffset == 0) && ip.__ipNode.isWhitespaceOnly)
		{
			ip.__backOne();
			ip.__ipOffset++;
			return ip;	
		}

		// three cases where must move forward:
		// - same_line && not whitespace but also not IPToken: move back to whitespace!
		// - at_top: didn't move so move forward as whitespace must be collapsed at block start
		// - crossed_block: move forward to valid character after seed whitespace
		if(((result == InsertionPoint.SAME_LINE) && !ip.whitespace && !ip.IPToken) ||
		   (result == InsertionPoint.AT_TOP) ||
		   (result == InsertionPoint.CROSSED_BLOCK))
		{
			ip.forwardOne();
		}
	}	
	
	return ip;
}

/*
 * Simple version of creation that doesn't check if the seed and seedOffset are valid. This is used for basic 
 * creation, forward and back testing.
 */
function __createInsertionPoint(top, seed, seedOffset)
{
	var ip = new InsertionPoint(top, seed, seedOffset, null, seed.parentNode.computedStyleValue("white-space"));
	return ip;
}

/**
Gets the "top" element ( __top ).
@return top
@type Element
*/
InsertionPoint.prototype.__defineGetter__(
	"top",
	function() {return this.__top;}
);

/**
 * POST05: check that ipnode is within top or is top!
 */
InsertionPoint.prototype.resetTop = function(newTop)
{
	this.__top = newTop;
}

InsertionPoint.prototype.__defineGetter__(
	"cssWhitespace",
	function() {return this.__cssWhitespace;}
);

/*
Does this IP have a "pre" value for the "whitespace" computed style value?

<p>Note: CSS whitespace values are : normal | pre | nowrap | pre-wrap | pre-line | inherit.
<p>"pre" and "pre-wrap" can be handled the same. Differnt handling will be required
to support "pre-line".

@returns true for "pre" or "pre-wrap"; false otherwise.
@type boolean 
*/
InsertionPoint.prototype.isPre = function()
{
	switch (this.__cssWhitespace) {
		case "pre":
		case "pre-wrap":
			return true;	
		default :
			return false;
	}
}

InsertionPoint.prototype.__defineGetter__(
	"ipNode",
	function() {return this.__ipNode;}
);

InsertionPoint.prototype.__defineGetter__(
	"ipOffset",
	function() {return this.__ipOffset;}
);

InsertionPoint.prototype.__defineSetter__(
	"ipOffset",
	function(value) {this.__ipOffset = value;}
);

/**
 * Get the offset of an insertion point with its line
 */
InsertionPoint.prototype.__defineGetter__(
	"lineOffset",
	function()
	{
		var line = this.line;
		var ipForCount = line.firstInsertionPoint;
		var count = 0;
		while(!ipForCount.equivalent(this))
		{
			count++;
			ipForCount.forwardOne();		
		}
		return count;
	}
);

/**
Gets the node referenced by this insertion point.
<ul>
<li>Returns ipNode if it is a Text Node.
<li>If __ipOffset points after the last node the last node is returned.
<li>If after _ipOffset is a text node return the previous (element?) node.
<li>Else return the (element?) node after _ipOffset.
</ul>

@return ipReferencedNode
@type Node
*/ 
InsertionPoint.prototype.__defineGetter__(
	"ipReferencedNode",
	function()
	{
		if(this.__ipNode.nodeType == Node.TEXT_NODE)
			return this.__ipNode;
		
		//else element reference: either right before or right after an element
		
		var offsetToUse;
		if( this.__ipOffset == this.__ipNode.childNodes.length ) {
		//if pointing after the last node, reference the last node.
			offsetToUse = this.__ipOffset-1;
		} else if ( this.__ipNode.childNodes[this.__ipOffset].nodeType == Node.TEXT_NODE ) {
		//else if the node after this point is a text node, reference the previous node.
			offsetToUse = this.__ipOffset-1;
		} else {
		//else reference the next node
			offsetToUse = this.__ipOffset;
		}
		return this.__ipNode.childNodes[offsetToUse];
	}
);

// Get rid of this???
/*
InsertionPoint.prototype.toTextReference = function()
{
	// if already point to text node then all is well
	if(this.ipNode.nodeType == Node.TEXT_NODE)
		return;

	// if point to empty element node then leave alone
	if((this.ipNode.childNodes.length != this.ipOffset)
	&& (this.ipNode.childNodes[this.ipOffset].nodeType == Node.ELEMENT_NODE)
	&& (this.ipNode.childNodes[this.ipOffset].contentType == Element.EMPTY_CONTENTTYPE))
		return;

	//MAX : BUG : if this.ipNode.childNodes.length == this.ipOffset what should we do???
	
	//Point to first
	if((this.ipNode.childNodes.length != this.ipOffset)
	&& (this.ipNode.childNodes[this.ipOffset].nodeType == Node.TEXT_NODE) {
		var newip = documentCreateInsertionPoint(this.top, this.ipNode.childNodes[this.ipOffset], 0);
		this.set(newip);
	}
}
*/

/**
 * Does the insertion point point to a whitespace?
 *
 * http://www.w3.org/TR/REC-CSS2/text.html#white-space-prop
 * "space" (Unicode code 32), "tab" (9), "line feed" (10), "carriage return" (13), and "form feed" (12)
 *
 * Note that this presumes white-space:pre
 */
InsertionPoint.prototype.__defineGetter__(
	"whitespace",
	function()
	{
		var character = this.character;

		if(character == null) // means on element node or at end of text node followed by element node
			return false;

		if(character == "") // means end of text node at end of line
			return true;

		return(eDOM.SPCHARS.indexOf(character) != -1);
	}
);

/**
Is this the start of the line?

@return true if start of line, false otherwise.
@type Boolean.
*/
InsertionPoint.prototype.__defineGetter__(
	"startOfLine",
	function()
	{
		var previp = this.clone();
		var prevResult = previp.backOne();
		if(prevResult != InsertionPoint.SAME_LINE)
			return true;
		return false;
	}
);
/**
Is this the end of the line?

@return true if end of line, false otherwise.
@type Boolean.
*/
InsertionPoint.prototype.__defineGetter__(
	"endOfLine",
	function()
	{
		var nextip = this.clone();
		var nextResult = nextip.forwardOne();
		if(nextResult != InsertionPoint.SAME_LINE)
			return true;
		return false;
	}
);

/**
 * IP tokens serve as the only non whitespace children of otherwise "ANY_CONTENTTTYPE" Elements. Without the token, Geiko
 * wouldn't render these nodes properly and an editor wouldn't know that it is valid to enter text within them. An editor
 * should replace a token when inserting content rather than just appending content around it.
 *
 * For XHTML, a IP token is an empty BR element OR an NBSP-only text node that are the only elements in their line.
 * 
 * POST05:
 * - add support for BR (move into eDOMHTML?) ie/ override this method with an equivalent ie/ copy its contents (specialization?)
 */
InsertionPoint.prototype.__defineGetter__(
	"IPToken",
	function()
	{
		// first case: NBSP on its own: POST05 - move this down to line as shouldn't presume what a token line is
		if((this.__ipNode.nodeType == Node.TEXT_NODE) && 
		   (this.__ipOffset < this.__ipNode.nodeValue.length) &&
		   (this.__ipNode.nodeValue.charAt(this.__ipOffset) == STRING_NBSP))
		{
			// See if more than this nbsp in the line (ie/ match non-nbsp whitespace) - if not then go before the NBSP
			var line = __createCSSLineFromNonBlockIP(this, false);
			if(line.tokenLine)
				return true;
		}
		return false;
	}
);

/**
 * Return line that insertion point is in
 */
InsertionPoint.prototype.__defineGetter__(
	"line",
	function()
	{
		return documentCreateCSSLine(this);
	}
);

/**
 * What character is at the point. Three cases:
 * - character at offset in referenced text node
 * - empty string "" if references the end of a text node that ends a line
 * - null if "character" is or is effectively an element
 *
 * Note: this assumes white-space of pre so end of line whitespace will be treated as valid characters rather
 * than as invisible collapsed markup
 */
InsertionPoint.prototype.__defineGetter__(
	"character",
	function()
	{
		// may be premature: could have text node ahead
		if(this.__ipNode.nodeType == Node.ELEMENT_NODE)
			return null;

		if(this.__ipOffset < this.__ipNode.length)
			return this.__ipNode.nodeValue.charAt(this.__ipOffset);

		var forwardip = this.clone();

		var result = forwardip.__forwardOne();

		var returnValue = ""; // at end of text node and forward is crossed_block or at_top then ""

		if(result == InsertionPoint.SAME_LINE)
		{
		//MAX - this goes wrong with br as an inline element.
		//Going forwardOne past the break gives ipNode as an Element Node, with a node offset from the line parent.
		//Either forwardOne should always return the next text node, if there is one, or we need to
		//test here if ipNode and offset point to a text node. 
			if(forwardip.__ipNode.nodeType == Node.ELEMENT_NODE)
				return null;

			returnValue = forwardip.__ipNode.nodeValue.charAt(0);
		}

		return returnValue;
	}
);

/**
 * One ip may directly reference a point in a text node while another may reference before or after a block or inline
 * empty element. They are equivalent if inserting text at this point leads to the same outcome.
 */
InsertionPoint.prototype.equivalent = function(ipToTest)
{
try {
	if(this.ipNode == ipToTest.ipNode) 
	{
		if(this.ipOffset == ipToTest.ipOffset)
			return true;
		return false;
	}
	
	// this refers to a text node: ipToTest has element offset
	if((this.ipNode.nodeType == Node.TEXT_NODE) && (ipToTest.ipNode.nodeType == Node.ELEMENT_NODE))
		return __elementReferenceEqualsTextReference(ipToTest.ipNode, ipToTest.ipOffset, this.ipNode, this.ipOffset);

	// this refers to element node: ipTo
	if((this.ipNode.nodeType == Node.ELEMENT_NODE) && (ipToTest.ipNode.nodeType == Node.TEXT_NODE))
		return __elementReferenceEqualsTextReference(this.ipNode, this.ipOffset, ipToTest.ipNode, ipToTest.ipOffset);
}
catch (e){
  return false;
}

	return false;
}

function __elementReferenceEqualsTextReference(el, elOffset, text, textOffset)
{
	// all comparisons must be against top inline ancestor: a/cs for Text and for <span>Text</span>
	var textILAncestor = text.topInlineAncestor;

	// element must be block ancestor of text node for an equivalent reference
	if(textILAncestor.parentNode != el)
		return false;

	// text ref points to end of text node and element reference points after text node
	if((textOffset == text.nodeValue.length) && (elOffset == (textILAncestor.offset + 1)))
		return true;

	// text ref points to start of text node and element reference points to text node
	if((textOffset == 0) && (elOffset == textILAncestor.offset))
		return true;

	return false;
}

InsertionPoint.prototype.clone = function()
{
	var clone = new InsertionPoint(this.__top, this.__ipNode, this.__ipOffset, this.__cw, this.__cssWhitespace);
	return clone;
}

InsertionPoint.prototype.set = function(ip)
{
	this.__top = ip.__top;
	this.__ipNode = ip.__ipNode;
	this.__ipOffset = ip.__ipOffset;
	this.__cw = ip.__cw;
	this.__cssWhitespace = ip.__cssWhitespace;
	this.__markedOffset = ip.__markedOffset;
}

/**
 * @returns the table cell element that contains this line or that is an ancestor of this line's container or null
 * if there is no such element.
 */
InsertionPoint.prototype.__defineGetter__(
	"tableCellAncestor",
	function()
	{
		var nodeToTest = this.ipNode.parentElement;
		while(nodeToTest != this.top)
		{
			if(nodeToTest.computedStyleValue("display") == "table-cell")
				return nodeToTest;
			nodeToTest = nodeToTest.parentNode;				
		}
		return null;
	}
);
/**
Sets this IP to the first Insertion Point.
*/
InsertionPoint.prototype.setToStart = function()
{
	var fip = this.__top.firstInsertionPoint(this.__top);
	if(fip)
		this.set(fip);
}

InsertionPoint.prototype.__backOne = function()
{
	var currentNode;
	var ipni, ipniResult;

	// special case 1: within a text node
	if(this.__ipNode.nodeType == Node.TEXT_NODE) {
		// in text node and not at the start then move back one in that node: assumption "pre" value won't change!
		if(this.__ipOffset > 0) {
			this.__ipOffset--;
			return InsertionPoint.SAME_LINE;
		}
		currentNode = this.__ipNode;
	}
	// special case 2: beyond last child (an empty element) in a container
	else if(this.__ipNode.childNodes.length == this.__ipOffset)
	{
		this.__ipOffset--;
		if(this.__ipNode.childNodes[this.__ipOffset].computedStyleValue("display") == "inline") {
		//If inline element, then still in same line
			ipni = new __IPNodeIterator(this.__top);
			ipni.currentNode = this.__ipNode.childNodes[this.__ipOffset];
			ipniResult = ipni.previousNode();
			if(!ipniResult && ipni.currentNode && (ipni.currentNode.nodeType == Node.TEXT_NODE)) {
			//If previous node in this line is a text node, point after the last character
				this.__ipNode = ipni.currentNode;
				this.__ipOffset = ipni.currentNode.nodeValue.length;
			}
			return InsertionPoint.SAME_LINE;		
		} else { // need to go to text node before empty element if can!
			ipni = new __IPNodeIterator(this.__top);
			ipni.currentNode = this.__ipNode.childNodes[this.__ipOffset];
			ipniResult = ipni.previousNode();
			if(!ipniResult && ipni.currentNode && (ipni.currentNode.nodeType == Node.TEXT_NODE)) {
				this.__ipNode = ipni.currentNode;
				this.__ipOffset = ipni.currentNode.nodeValue.length;
			}
			return InsertionPoint.CROSSED_BLOCK;
		}
	}
	else {
		currentNode = this.__ipNode.childNodes[this.__ipOffset];
	}

	// Must move back to previous editable node	
	ipni = new __IPNodeIterator(this.__top);
	ipni.currentNode = currentNode;
	var ipniBlockCrossed = ipni.previousNode();
	if(ipni.currentNode == null) // at top - no change to IPTYPE
		return InsertionPoint.AT_TOP

	// may have crossed to new parent: reset cssWhitespace setting
	this.__cssWhitespace = ipni.currentNode.parentNode.computedStyleValue("white-space");

	if(ipni.currentNode.nodeType == Node.TEXT_NODE)
	{
		this.__ipNode = ipni.currentNode;
		this.__ipOffset = this.__ipNode.nodeValue.length; 

		if(ipniBlockCrossed)
			return InsertionPoint.CROSSED_BLOCK;

		this.__ipOffset--; // back one if same line
		return InsertionPoint.SAME_LINE;
	}
	
	// currentNode is empty element
	this.__ipNode = ipni.currentNode.parentNode;
	if(ipniBlockCrossed) {
		// special case: if block level node is the empty element then jump before it
		if(ipniBlockCrossed == ipni.currentNode) {
			// special case: need to reach ahead into a text node if ipniResult and it is the current node. That may
			// expose whitespace and we don't want that
			currentNode = ipni.currentNode;
			ipniResult = ipni.previousNode();
			if(!ipniResult && ipni.currentNode && (ipni.currentNode.nodeType == Node.TEXT_NODE)) {
				this.__ipNode = ipni.currentNode;
				this.__ipOffset = ipni.currentNode.nodeValue.length;
			}
			else {
				this.__ipOffset = currentNode.offset;
			}
		}
		else {		
			// if block crossed and now on a different empty element then go after it
			this.__ipOffset = ipni.currentNode.offset + 1; // after empty element - default is block
		}
		return InsertionPoint.CROSSED_BLOCK;
	}

	this.__ipOffset = ipni.currentNode.offset;
	return InsertionPoint.SAME_LINE;
}

/**
 * Assume the InsertionPoint is created in a valid position.
 *
 * POST05:
 * - account for linefeed right before a tag. May actually do in "__backOne"
 */
InsertionPoint.prototype.backOne = function()
{
	var origip = this.clone(); // jump back here if only whitespace between current position and top

	var result = this.__backOne();

	if(result == InsertionPoint.AT_TOP)
		return InsertionPoint.AT_TOP;

	// Collapsed whitespace: MAY have to jump back if in a non pre text node, either at its end or on a whitespace
	if((this.__ipNode.nodeType == Node.TEXT_NODE) && (this.__cssWhitespace != "pre") &&
	   ((this.__ipNode.nodeValue.length == this.__ipOffset) ||
	    (eDOM.SPCHARS.indexOf(this.__ipNode.nodeValue.charAt(this.__ipOffset)) != -1)))
	{			
		// Now on whitespace in TEXT_NODE: go back until get to non whitespace
		var backip = this.clone();
		var firstwspip;
		var resultBack;
		do
		{
			firstwspip = backip.clone();

			resultBack = backip.__backOne();

			if(resultBack != InsertionPoint.SAME_LINE)
				result = resultBack;

			if(resultBack == InsertionPoint.AT_TOP)
				break;
		}
		// text node and not pre and whitespace (end of text node or literally a whitespace character)
		while((backip.__ipNode.nodeType == Node.TEXT_NODE) && (backip.__cssWhitespace != "pre") &&
		      ((backip.__ipNode.nodeValue.length == backip.__ipOffset) ||
		       (eDOM.SPCHARS.indexOf(backip.__ipNode.nodeValue.charAt(backip.__ipOffset)) != -1)))	

		if(result == InsertionPoint.CROSSED_BLOCK) // CROSSED_BLOCK: adopt when backcp takes you
		{
			// Keep the place that backcp went to
			this.set(backip);

			// Jump right after the non whitespace unless already there (resultBack is CROSSED_BLOCK if line ends in Element)
			// or this now points to an editable token
			if(!((resultBack == InsertionPoint.CROSSED_BLOCK) || this.IPToken))
				this.__ipOffset++;
		}
		else if(result == InsertionPoint.AT_TOP) // AT_TOP: return to original cp position
			this.set(origip);
		else // SAME_LINE: go back to the first whitespace in the sequence
		{
			if(backip.IPToken) // only for create in wrong place!
				this.set(backip);
			else
				this.set(firstwspip);
		}
	}
	this.__markedOffset = -1; // back or forward reset any markers
	return result;
}

/**
Move this Insertion Point forward by one.
Follows behavior of white-space:pre.

@return SAME_LINE | CROSSED_BLOCK | AT_TOP
@type integer
*/
InsertionPoint.prototype.__forwardOne = function()
{
	var textNodeAtEnd = null;
	//If TEXT_NODE and not pointing after the last character;
	if(this.__ipNode.nodeType == Node.TEXT_NODE) {
		if(this.__ipOffset < this.__ipNode.length) {
			//Increment offset to point to next character.
			this.__ipOffset++;
			return InsertionPoint.SAME_LINE;
		}
		else {
			textNodeAtEnd = this.__ipNode;
		}
	}

	//Either at end of text node, or before/after an Element node...
	var currentNode = this.__ipNode;

	if( currentNode.nodeType == Node.ELEMENT_NODE )	{
		var numChildren = currentNode.childNodes.length;
		//IF element, of EMPTY content type, need to point after it.
		if ( this.__ipOffset < numChildren ) {
			if ( currentNode.childNodes[this.__ipOffset].nodeType == Node.ELEMENT_NODE ) {
				if ( currentNode.childNodes[this.__ipOffset].contentType == Element.EMPTY_CONTENTTYPE ) {
					this.__ipOffset++;
					return InsertionPoint.SAME_LINE;
				}
			}
		}
		
		if ( this.__ipOffset < numChildren ) {
		//If before the last node;
			currentNode = this.__ipNode.childNodes[this.__ipOffset];
			if(currentNode.nodeType == Node.TEXT_NODE) {
			//If this is a text node;
				this.__ipNode = currentNode; 
				this.__ipOffset = 1;
				return InsertionPoint.SAME_LINE;
			}
		} else {
		//Else pointing after the last node:
		// need to set current node reference to before the last node. ????
		// ??? if element point before it, eif text node point after last char??? 
			var lastNode = this.__ipNode.childNodes[currentNode.childNodes.length-1];
			currentNode = lastNode;
		}
	}

	// Move forward to next editable node:	
	var ipni = new __IPNodeIterator(this.__top);
	ipni.currentNode = currentNode;
	var ipniResult = ipni.nextNode();

	//If not a next node, must be AT_TOP.
	if(ipni.currentNode == null) // at top!
		return InsertionPoint.AT_TOP;

	// May have crossed to new parent: reset cssWhitespace setting.
	this.__cssWhitespace = ipni.currentNode.parentNode.computedStyleValue("white-space");
		
	if(ipni.currentNode.nodeType == Node.TEXT_NODE) {
	//If next node is a Text Node;
		this.__ipNode = ipni.currentNode;
		if(ipniResult) {
		//If crossed to new line;
			this.__ipOffset = 0; 
			return InsertionPoint.CROSSED_BLOCK;
		}
		else {
		//Else same line;
			if ( textNodeAtEnd ) {
			//If crossing from one text node to another move ahead one place.
				this.__ipOffset = 1; //Is this always valid???
			} else {
				this.__ipOffset = 0; //go to first
			}
			return InsertionPoint.SAME_LINE;
		}
	}
	//Else Element Node;
	// empty (always inline if get this far) element
	currentNode = ipni.currentNode;
	this.__ipNode = currentNode.parentNode;

	if(ipniResult) {
	//If we've crossed to a new line
		if(ipniResult == ipni.currentNode) {
		// if new line starts with an empty element ie/ block crossed is the empty element
			// special case: need to reach ahead into a text node if ipniResult and it is the current node. That may
			// expose whitespace and we don't want that
			ipniResult = ipni.nextNode();
			if(!ipniResult && ipni.currentNode && (ipni.currentNode.nodeType == Node.TEXT_NODE)) {
				this.__ipNode = ipni.currentNode;
				this.__ipOffset = 0;
			} else {
				this.__ipOffset = currentNode.offset;
			}
		} else {
			this.__ipOffset = currentNode.offset;
		}
		return InsertionPoint.CROSSED_BLOCK;
	}
	else {
		if ( textNodeAtEnd ) {
			this.__ipOffset = currentNode.offset+1;
		} else {		
			this.__ipOffset = currentNode.offset;
		}
		return InsertionPoint.SAME_LINE;
	}
}

/**
 * Accounts for display setting of the block, top and the whitespace property. The whitespace property is testimont
 * to hand coding. As you can style lines properly with tagging, the need to accept text formatting characters as
 * more than whitespace is a throwback to hand coded HTML and limited ability to style tagged text.
 *
 * POST05:
 * - add skips for newlines right after or before any tag (http://www.w3.org/TR/REC-html40/appendix/notes.html#notes-line-breaks)
 */
InsertionPoint.prototype.forwardOne = function()
{
	var origip = this.clone(); // jump back here if only whitespace between current position and top

	// Special case: skip any editable token completely and move to the next line or stay put if at top
	var result;
	if(origip.IPToken)
	{
		do {
			result = this.__forwardOne();
		}
		while(result == InsertionPoint.SAME_LINE)

		if(result == InsertionPoint.AT_TOP) // revert to original position if AT_TOP
			this.set(origip);
		return result;
	}

	result = this.__forwardOne();

	// at top - no need to chase whitespace
	if(result == InsertionPoint.AT_TOP)
		return result;

	// Collapsed whitespace: if now on a non pre text node on whitespace and either original was whitespace OR
	// original was after an empty element that isn't succeeded by another element then must eat whitespace 
 	// and get to a non whitespace character or element
	if(((this.__cssWhitespace != "pre") && this.whitespace) &&
	   (origip.whitespace || (result == InsertionPoint.CROSSED_BLOCK) ||
            ((origip.__ipNode.nodeType == Node.ELEMENT_NODE) && ((origip.__ipNode.childNodes.length == origip.__ipOffset) || (origip.__ipNode.childNodes[origip.__ipOffset].nodeType != Node.ELEMENT_NODE)))))
 	{
		var resultForward;	
		do
		{
			resultForward = this.__forwardOne();

			if(resultForward != InsertionPoint.SAME_LINE)
				result = resultForward;
		}
		while((this.__cssWhitespace != "pre") && this.whitespace && (result != InsertionPoint.AT_TOP))

		// AT_TOP: return to original cp position
		if(result == InsertionPoint.AT_TOP) 
			this.set(origip);
		else if(result == InsertionPoint.CROSSED_BLOCK)
		{
			// Hate this but special case: leading collapsed whitespace of a line you cross into
			if((this.__cssWhitespace != "pre") && 
		           (this.__ipNode.nodeType == Node.TEXT_NODE) && 
			   (this.__ipNode.nodeValue.length == this.__ipOffset))
			{
				// move forward one and then reduce offset by 1
				this.__forwardOne();
				this.__ipOffset--;
			}
		}
	}

	this.__markedOffset = -1; // back or forward reset any markers

	return result;
}

/**
Go to same offset in previous line or end of previous line whichever is smaller.
*/
InsertionPoint.prototype.upOne = function()
{
	var previousLine = this.line.previousLine;

	if(!previousLine)
		return false;

	if(this.__markedOffset == -1)
		this.__markedOffset = this.lineOffset;

	var ip = previousLine.insertionPointAt(this.__markedOffset);

	this.set(ip);

	return true;
}

/**
Go to same offset in next line or end of previous line whichever is smaller.
*/
InsertionPoint.prototype.downOne = function()
{
	var nextLine = this.line.nextLine;

	if(!nextLine)
		return false;

	if(this.__markedOffset == -1)
		this.__markedOffset = this.lineOffset;

	var ip = nextLine.insertionPointAt(this.__markedOffset);

	this.set(ip);

	return true;
}
/**
Sets this Insertion Point to the last-IP of the line this Insertion Point is within.
*/
InsertionPoint.prototype.toLineEnd = function()
{
	this.set(this.line.lastInsertionPoint);
}

/**
Sets this Insertion Point to the first-IP of the line this Insertion Point is within.
*/
InsertionPoint.prototype.toLineStart = function()
{
	this.set(this.line.firstInsertionPoint);
}

/**
Sets this Insertion Point to the first-IP within the "top" Element for this Insertion Point.
*/
InsertionPoint.prototype.toTopStart = function()
{
	this.set(this.top.firstInsertionPoint(this.top));
}

/**
Sets this Insertion Point to the last-IP within the "top" Element for this Insertion Point.
*/
InsertionPoint.prototype.toTopEnd = function()
{
	this.set(this.top.lastInsertionPoint(this.top));
}

/** 
 * Delete a character or empty inline element that preceeds an insertion point. If only one character or element
 * in the line then this turns the line into a token line
 *
 * This is used as part of a more complex and editor specific "delete" operation
 *
 * @return false if at start of line
 */
InsertionPoint.prototype.deletePreviousInLine = function()
{
	// if at start => merge
	var startip = this.clone();
	var startResult = startip.backOne();

	// Special case one: start of line/at_top
	if(startResult != InsertionPoint.SAME_LINE)
		return false;

	// special case only element: check if only element - if it is then turn into ipToken line
	var previp = startip.clone();
	var prevResult = previp.backOne();
	var nextip = this.clone();
	var nextResult = nextip.forwardOne();
	if((prevResult != InsertionPoint.SAME_LINE) && (nextResult != InsertionPoint.SAME_LINE))
	{
		var line = this.line.deleteContents();
		this.set(line.firstInsertionPoint);
		return true;	
	}

	// special case eol wsp exposure: if deletion would expose a space then make sure that space becomes an nbsp
	if(nextResult != InsertionPoint.SAME_LINE)
	{
		if(previp.whitespace) // assuming previp in same line as checked one element line already
			previp.__ipNode.replaceData(previp.ipOffset, 1, STRING_NBSP);		
	}
	// special case wsp collapse: if deletion would collapse a whitespace, no matter where it would be
	// exposed then preserve it with an NBSP
	else 
	{
		if(this.whitespace && ((prevResult != InsertionPoint.SAME_LINE) || previp.whitespace))
		{
			// if whitespace actually starts following text node
			if(this.__ipNode.nodeValue.length == this.__ipOffset)
				nextip.__ipNode.replaceData(0, 1, STRING_NBSP);
			else
				this.__ipNode.replaceData(this.__ipOffset, 1, STRING_NBSP);
		}
	}

	// majority case: delete one character and any exclusive parents it has - restore to after the character
	// before this character
	var range = document.createRange();
	range.setStart(startip.ipNode, startip.ipOffset);
	range.setEnd(this.__ipNode, this.__ipOffset);
	var keepRange = range.cloneRange()

	var cssr = __createCSSRange(range, this.__top, false)

	cssr.includeExclusiveParents(); 

	// TMP: POST05: include exclusive shouldn't do a partial grab of one end or the other of a range but it does now!
	if((cssr.startContainer == startip.ipNode) || (cssr.endContainer == this.ipNode))
		keepRange.deleteContents();
	else
		cssr.deleteContents();

	this.__top.normalize();

	// restore
	if(prevResult != InsertionPoint.AT_TOP)
	{
		previp.forwardOne();
		this.set(previp);
	}
	else
		this.setToStart();

	return true;
}

/**
Can only split a contained line. 

POST05: a/c for pre element reference!
*/
InsertionPoint.prototype.splitContainedLine = function()
{
	var line = this.line;
	if(line.lineType == CSSLine.BOUNDED_LINE)
		return; // exception later?

	if(line.container == this.top)
		return; // exception later?

	if(line.container == line.tableCellAncestor)
		return; // exception later?

	if( this.isPre() )
	{
		// Mozilla bug: the line feed isn't treated properly by the renderer. If select the offset
		// after the linefeed, you end up on the same line as the line and suddenly jump once you
		// start inserting more characters. There is a mismatch in the offset setting of Selection
		// and the rendering of the selection. There is no workaround.
		// POST05: need to enter bug in Bugzilla
		// MAX : The browser manages to step correctly through PRE newlines so there must be some
		// way round this problem. Inserting whitespace after the newline helps a bit, but means you have an
		// extra char to move through or delete.
		// (maybe let browser handle left/right key press if a newline, otherwise let Mozile handle it.)
		this.__ipNode.insertData((this.__ipOffset), eDOM.LINEFEED); // if pre - insert linefeed
		this.__ipOffset++;		
//		this.__ipNode.insertData((this.__ipOffset), "\uFEFF"); // if pre - insert zero width no-break space.
//		this.__ipOffset++;		
		return;
	}

	var newLineContainer = line.container.cloneNode(false);

	// if at end of line then create an empty line after this one and move to it
	if(this.equivalent(line.lastInsertionPoint))
	{
		var emptyLineToken = eDOM.token();
		newLineContainer.appendChild(emptyLineToken);
		line.container.parentNode.insertAfter(newLineContainer, line.container);
		var newIP = this.clone();
		newIP.forwardOne();
		this.set(newIP);
		return;
	}

	// If split at start of line then create a new empty line before this line but leave the CP where it is
	if(this.equivalent(line.firstInsertionPoint))
	{
		newLineContainer.appendChild(eDOM.token());
		line.container.parentNode.insertBefore(newLineContainer, line.container);
		return;
	}

	// make a new line after this one with the end!
	var newLineRange = document.createRange();
	newLineRange.selectNodeContents(line.container);
	// issue - don't want to grab 
	newLineRange.setStart(this.ipNode, this.ipOffset);
	var newLineContents = newLineRange.extractContents();
	line = documentCreateCSSLine(line.firstInsertionPoint);
	line.normalizeWhitespace(); // takes care of rubbish at start or end ie/ part of invalid styled text
	newLineContainer.appendChild(newLineContents);
	line.container.parentNode.insertAfter(newLineContainer, line.container);
	var newLine = documentCreateCSSLine(newLineContainer.firstInsertionPoint(line.top));
	newLine.normalizeWhitespace(); // takes care of rubbish at start or end ie/ part of invalid styled text
	newLine.forceLineBreaksBeforeAfter();
	this.set(newLine.firstInsertionPoint);
}

/**
Insert a character into a line.
 
POST05:
- fix to check if next inline element is a text node and insert there to a/c for span after object: use clone.forwardOne
- mix with insertTextNode

@param {Integer} charCode
*/
InsertionPoint.prototype.insertCharacter = function(charCode)
{
	// Simple - experiment - doesn't account for tokens [still need whitespace tests in most cases!]
	if(this.__ipNode.nodeType == Node.ELEMENT_NODE)
	{
		// after element node at end of line with container or editable area: insert after
		if(this.__ipOffset == this.__ipNode.childNodes.length)
		{
			this.__ipNode.appendChild(document.createTextNode(""));
			this.__ipNode = this.__ipNode.lastChild;
			this.__ipOffset = 0;
		}
		else
		{
			var referencedNode = this.__ipNode.childNodes[this.__ipOffset];
			
			// ie/ between element node and text node: ie/ after element node
			if(referencedNode.nodeType == Node.TEXT_NODE)
			{
				this.__ipNode = referencedNode;
				this.__ipOffset = 0;
			}
			// references element node: either before element node OR between element nodes
			else // [Bug: insert after if before block level node ie/ end of containerless line!]
			{
				// ie/ between text node and element node: ie/ before element node with text node before it
				if(referencedNode.previousSibling && (referencedNode.previousSibling.nodeType == Node.TEXT_NODE))
				{
					this.__ipNode = referencedNode.previousSibling;
					this.__ipOffset = this.__ipNode.nodeValue.length;
				}
				// no text node before the element node - insert one
				else
				{
					var emptyTextNode = document.createTextNode("");
					this.__ipNode.insertBefore(emptyTextNode, referencedNode);
					this.__ipNode = emptyTextNode;
					this.__ipOffset = 0;
				}				
			}			
		}
	}
	// lot's of whitespace/nbsp handling if not dealing with "pre"
	else if( !this.isPre() )
	{
		// TODO: problem for CHARCODE_SPACE - if only textToken ... node.textToken should be a first class check
		var prevCharPointer;
		// Cases: start line, end line, space is current character or space before
		if(charCode == CHARCODE_SPACE) 
		{
			// whitespace ahead or end of line
			if(this.whitespace)
			{
				charCode = CHARCODE_NBSP;
			}
			// check if previous character is a space or at beginning of line
			else 
			{
				prevCharPointer = this.clone();
				// beginning of line if going back brings us over a block boundary
				if((prevCharPointer.backOne() != InsertionPoint.SAME_LINE) || prevCharPointer.whitespace)
					charCode = CHARCODE_NBSP;	
			}
		}	
		// add non whitespace - switch a preceding nbsp if it is in same line and isn't preceded by a
    // space or nbsp (reason for leaving sequences of nbsp's is that it makes dealing with at top situations
		// easier.	
		else 
		{ 
			prevCharPointer = this.clone();
			var prevResult = prevCharPointer.backOne();
			if((prevResult == InsertionPoint.SAME_LINE) && (prevCharPointer.character == STRING_NBSP))
			{
				var tnToReplace = prevCharPointer.ipNode;
				var tnOffset = prevCharPointer.ipOffset;
				if((prevCharPointer.backOne() == InsertionPoint.SAME_LINE) && !prevCharPointer.whitespace)
					tnToReplace.replaceData(tnOffset, 1, STRING_SPACE);
			}

			// TODO: use node.textToken here

			// if we're at the first point in a line and nbsp is the only character then replace nbsp with a character
			if((this.clone().backOne() != InsertionPoint.SAME_LINE) && (this.character == STRING_NBSP)) // start line
			{
				var next = this.clone();
				var result = next.forwardOne(); // beyond nbsp
				if(result != InsertionPoint.SAME_LINE)
				{
					this.__ipNode.replaceData(this.__ipOffset, 1, String.fromCharCode(charCode));
					this.__ipOffset++;
					return;	
				}
			}
		}	
	}
	this.__ipNode.insertData(this.__ipOffset, String.fromCharCode(charCode));
	this.__ipOffset++;
}

/**
 * Insert a node
 *
 * Valid nodes are:
 * - text nodes
 * - empty elements (block or inline) ex/ BR, HR or IMG in XHTML
 * - other elements (any but those with display list-item or table-cell)
 * If the current position is in a contained line and the node being inserted is a block
 * or table element then the line will be split and the node inserted between the two 
 * halves. Otherwise the node is inserted within the line.
 *
 * IP is reset to first position within, or for empty elements after, the node.
 * 
 * POST05: a/c for invisible or visible whitespace (make into NBSP if necessary)
 */
InsertionPoint.prototype.insertNode = function(node)
{
	if(node.nodeType == Node.TEXT_NODE) {
		this.__insertTextNode(node);
		return;
	}

	// assume element node - if not then exception
	if(node.nodeType != Node.ELEMENT_NODE)
		return; // POST05: exception

	// empty element
	if(node.contentType == Element.EMPTY_CONTENTTYPE) {
		this.__insertElement(node);
		return;
	}

	var nodeDisplayValue = node.computedStyleValue("display");
	if(nodeDisplayValue == "inline") { // inline element
		this.__insertElement(node);
		return;
	}

	// can only insert blocks or tables
	if(!((nodeDisplayValue == "block") || (nodeDisplayValue == "table")))
		return; // EXCEPTION POST05

	var line = this.line;
	if(line.lineType == CSSLine.BOUNDED_LINE) {
		this.__insertElement(node);
		return;
	}
	//Else CONTAINED_LINE...

	// treat a list or table line just like a bounded line: POST05 - treat these better?
	if(line.containedLineType != ContainedLine.BLOCK) {
		this.__insertElement(node);
		return;
	}

	// check if at start or end of line
	if(this.equivalent(line.firstInsertionPoint))
		line.container.parentNode.insertBefore(node, line.container);
	else if(this.equivalent(line.lastInsertionPoint))
		line.container.parentNode.insertAfter(node, line.container);
	else {
		// split contained line
		this.splitContainedLine();
		line.container.parentNode.insertAfter(node, line.container);
	}
	var newip = node.firstInsertionPoint(this.top);
	this.set(newip);
}	

/**
 * Insert an element into a line
 */
InsertionPoint.prototype.__insertElement = function(element)
{
	if(this.ipNode.nodeType == Node.ELEMENT_NODE) {
		if(this.ipNode.childNodes.length == this.ipOffset)
			this.ipNode.appendChild(element);
		else
			this.ipNode.childNodes[this.ipOffset].parentNode.insertBefore(element, this.ipNode.childNodes[this.ipOffset]);
	} else {	// text ip
		var nextip = this.clone();
		nextip.forwardOne();
		var previp = this.clone();
		previp.backOne();

		var inlineParent = null;
		if(this.ipNode.parentNode.computedStyleValue("display") == "inline")
			inlineParent = this.ipNode.parentNode;

		if(nextip.ipNode != this.ipNode) { //Bug when at end of only line in editable section.?
		// at end of text node => nothing to return ...
			if(inlineParent)
				inlineParent.parentNode.insertAfter(element, inlineParent);
			else
				this.ipNode.parentNode.insertAfter(element, this.ipNode);
		}
		else if(previp.ipNode != this.ipNode) {
		// at beginning of text node
			if(inlineParent)
				inlineParent.parentNode.insertBefore(element, inlineParent);
			else
				this.ipNode.parentNode.insertBefore(element, this.ipNode);
		}
		else {
		// not at either end of text node, so within the text.
			var secondTextNode = this.ipNode.splitText(this.ipOffset);
			if(inlineParent) {
				var newInlineContainer = inlineParent.cloneNode(false);
				inlineParent.parentNode.insertAfter(newInlineContainer, inlineParent);
				newInlineContainer.appendChild(secondTextNode);
				newInlineContainer.parentNode.insertBefore(element, newInlineContainer);
			}
			else {
				this.ipNode.parentNode.insertBefore(element, secondTextNode);
			}
		}
	}

	var newip = element.firstInsertionPoint(this.top);
	this.set(newip);
}

/**
Insert a text node at ipNode, ipOffset.
Update this IP to point after(?) the inserted node.
*/
InsertionPoint.prototype.__insertTextNode = function(text)
{
	var insertedNode;
	if(this.ipNode.nodeType == Node.TEXT_NODE) {
		var parentNode = this.ipNode.parentNode;
		if ( this.ipOffset == 0 ) {
			insertedNode = parentNode.insertBefore(text, this.ipNode);
		} else if ( this.ipOffset >= this.ipNode.length ) {
			insertedNode = parentnode.insertAfter(text, this.ipNode);
		} else {
			var node2 = this.ipNode.splitText(this.ipOffset);
			insertedNode = parentNode.insertBefore(text, node2);
		}
	}
	else {
		// element reference
		if(this.ipNode.childNodes.length == this.ipOffset) {
			insertedNode = this.ipNode.appendChild(text);
		} else { 
			insertedNode = this.ipNode.insertBefore(text, this.ipNode.childNodes[this.ipOffset]);
		}
	}

	var newip = documentCreateInsertionPoint(this.top, insertedNode, insertedNode.nodeValue.length);
	this.set(newip);
}

/**
 * Insertion point node iterator.
 *
 * Iterates through three types of Insertion Point Node that lie within a specific container or scope:
 * - non empty TextNodes: text nodes with one non whitespace or nbsp character
 * - inline empty element
 * - block level empty element
 *
 * Next/previous return true if iteration crosses or happens on a block level element. They
 * return false otherwise.
 *
 * The iterator is initialized on the first InsertionPoint node within the iterators scope
 *
 * currentNode contains the current InsertionPoint node of the iterator. It is null if the insertionPoint
 * is placed before or beyond valid insertion points within its scope.
 */
function __IPNodeFilter()
{
	this.__crossedBlock = null;
}

__IPNodeFilter.prototype.__defineGetter__(
	"crossedBlock",
	function() {return this.__crossedBlock;}
);

__IPNodeFilter.prototype.reset = function()
{
	this.__crossedBlock = null;
}

/**
 * Accept any node that could be selected as an insertion point when white-space is "pre" ie/ the most liberal definition.
 * This filter (like most of eDOM) is driven by CSS and not other indicators of whitespace such as XML attributes
 * (see: http://www.w3.org/TR/2000/REC-xml-20001006#sec-white-space)
 *
 * POST05: 
 * - consider returning AT_TOP, CROSSED_LINE, SAME_LINE from next and previous ala InsertionPoint (easier currentNode == null means top etc)
 * - search down the preceding or succeeding inline element to make sure it has a selectable child
 * - consider redoing special case text handling: put in memory of where we are: if in line then 
 */
__IPNodeFilter.prototype.acceptNode = function(node)
{
	if(node.nodeType == Node.TEXT_NODE)
	{
		// empty node - not visible
		if(node.nodeValue.length == 0)
			return NodeFilter.FILTER_REJECT;

		// if non NBSP whitespace then normally reject - except (special cases) if it is before/after an inline node OR it is
		// within an inline node
		if(!(/\u00A0+/.test(node.nodeValue)) && node.isWhitespaceOnly)
		{
			// accept empty text node that is within an inline element
			if(node.parentNode.computedStyleValue("display") == "inline")
				return NodeFilter.FILTER_ACCEPT;

			// accept empty text that has either an inline node before or after it: strictly speaking should
			// make sure that the inline node has a text node or empty object as a child!
			if((node.previousSibling && (node.previousSibling.nodeType == Node.ELEMENT_NODE) && (node.previousSibling.computedStyleValue("display") == "inline")) &&
			   (node.nextSibling && (node.nextSibling.nodeType == Node.ELEMENT_NODE) && (node.nextSibling.computedStyleValue("display") == "inline")))
				return NodeFilter.FILTER_ACCEPT;

			return NodeFilter.FILTER_REJECT;
		}
			
		return NodeFilter.FILTER_ACCEPT;
	}

	// normally we skip elements: two cases are of interest - when we cross a block and when we find an
	// empty inline element. We need to note when we are in a new line and empty inline elements are 
	// selectable. However, we skip all other types of element including childless, child bearing elements 
	// as Geiko doesn't display these. eDOM assumes that they are just for spacing and should not support insertion.
	if(node.nodeType == Node.ELEMENT_NODE) {
		if(node.computedStyleValue("display") != "inline") {
			if(!this.__crossedBlock)
				this.__crossedBlock = node;
		}
		if(node.contentType == Element.EMPTY_CONTENTTYPE) {
		// allows empty block and inline elements
			return NodeFilter.FILTER_ACCEPT;
		}
	}

	return NodeFilter.FILTER_SKIP;
}

/**
 * By default the current node is set to the first child under root that passes the filter or null if there are
 * no nodes to iterate through
 */
function __IPNodeIterator(root)
{
	this.__ipNodeFilter = new __IPNodeFilter();

	this.__root = root;

	this.__ipNodeWalker = document.createTreeWalker(root,
					NodeFilter.SHOW_ALL,
					this.__ipNodeFilter,
					false);

	this.__currentNode = this.__ipNodeWalker.firstChild();
}

__IPNodeIterator.prototype.setToStart = function()
{
	this.__ipNodeWalker.currentNode = this.__root;
	this.__currentNode = this.__ipNodeWalker.firstChild();
	return this.__currentNode;
}

__IPNodeIterator.prototype.setToEnd = function()
{
	this.__ipNodeWalker.currentNode = this.__root;
	this.__currentNode = this.__ipNodeWalker.lastChild();
	return this.__currentNode;
}

__IPNodeIterator.prototype.__defineGetter__(
	"currentNode",
	function() {return this.__currentNode;}
);

__IPNodeIterator.prototype.__defineSetter__(
	"currentNode",
	function(value) {this.__currentNode = value; this.__ipNodeWalker.currentNode = value;}
);

__IPNodeIterator.prototype.nextNode = function()
{
	this.__ipNodeFilter.reset();
	var oldBlockParentNode = this.__currentNode ? this.__currentNode.__nilParentElement : null;
	this.__currentNode = this.__ipNodeWalker.nextNode(); 
	// check special case: if cross to new bounded line from existing contained line then "real" block
	// is the container of the contained line. Tree skips that naturally so must a/c for it. Two situations - 
	// didn't cross a block according to the tree so let's check and now on the "block", the empty element. Is
	// this really the only block?
	if(!this.__ipNodeFilter.crossedBlock || (this.__currentNode && (this.__ipNodeFilter.crossedBlock == this.__currentNode)))
	{
		var newBlockParentNode = this.__currentNode ? this.__currentNode.__nilParentElement: null;
		if(newBlockParentNode != oldBlockParentNode)
			return newBlockParentNode;
	}
	return this.__ipNodeFilter.crossedBlock;
}

__IPNodeIterator.prototype.previousNode = function()
{
	this.__ipNodeFilter.reset();

	this.__currentNode = this.__ipNodeWalker.previousNode(); 

	return this.__ipNodeFilter.crossedBlock;
}

/*************************************************************************************************************
 * CSSLine captures the concept of a line. Lines may be bounded or contained. 
 *
 * IP establishes a valid seed for a line:
 * - before block level empty element (means will be bounded)
 * - after block level element (means bounded - but need not be empty)
 * - before inline element or within text (means creates non empty bounded or a contained line)
 *
 * TODO:
 * - Element.lines (ala Range.lines)
 * - line.removeCollapsedWhitespace()
 * - line.newLineBeforeAfter()
 * - insertElementAfter?
 * 
 * POST05:
 * - handle previous/next ie/ up/down 
 * - handle line length and offset
 * - handle destroyed line ie/ collapsed range - must throw exceptions ...
 * - add style to this: ie/ styleLine == styleElement with container application if 
 * necessary
 * - utility: normalizeWhiteSpace removes white space before the beginning or after the end of lines. Note that this 
 * should be applied carefully because it will effect TextPointers. It will also merge any a sequence of white
 * spaces into one space and remove white space at the beginning or end of the line. This should remove the
 * need for a lot of special case handling of white spaces. It should also simplify post split behaviors.
 * - utility: formatLineMarkup - does normalize of whitespace AND sets container on its own line
 * - ** incorporate validation support ie/ dom level 3 - can split line? ie/ delegates to container.nextSibling check
 *   ditto for "delete"/merge lines. 
 *************************************************************************************************************/

Document.prototype.createCSSLine = documentCreateCSSLine;

/**
 * IP establishes a valid seed for a line:
 * - before block level empty element (means will be bounded)
 * - after block level element (means bounded - but need not be empty)
 * - before inline element or within text (means creates non empty bounded or a contained line)
 */
function documentCreateCSSLine(ip)
{
	if(ip.ipNode.nodeType == Node.TEXT_NODE)
		return __createCSSLineFromNonBlockIP(ip, true);
	
	var seed = ip.ipReferencedNode;
	if(seed.computedStyleValue("display") == "inline")
		return __createCSSLineFromNonBlockIP(ip, true);

	// special handling
	return __createBoundedLineFromBlockIP(ip, true);
}

/**
 * Create a bounded line from an ip that references a block level element
 *
 * - either after an element node or before one (may be both in which case, treated as if before one!)
 * - block element is a boundary and not part of a line per se
 * - always creates a bounded line and that line may be empty: in this case, startip == endip == seed ip
 */
function __createBoundedLineFromBlockIP(ip, handleCollapsedWhitespace)
{
	var seed = ip.ipReferencedNode;
	var startBoundary = null;
	var endBoundary = null;
	var startip;
	var endip;
	var elementCount = 0;
	var ipNodeIterator = new __IPNodeIterator(ip.top);

	// if before a block element then this is the end node
	if(seed.offset == ip.ipOffset)
	{
		endBoundary = seed;
		endip = ip.clone();
		var startIPNode = null;
		ipNodeIterator.currentNode = seed;

		// go back while not at top and in same line
		while(!ipNodeIterator.previousNode() && ipNodeIterator.currentNode)
		{
			startIPNode = ipNodeIterator.currentNode;
			if(startIPNode.nodeType == Node.ELEMENT_NODE)
				elementCount++;
		}

		// at least one node is within the line!
		if(startIPNode)
		{ 
			startBoundary = startIPNode.topInlineAncestor.previousSibling;
			if(startIPNode.nodeType == Node.ELEMENT_NODE)
				startip = handleCollapsedWhitespace ? documentCreateInsertionPoint(ip.top, startIPNode.parentNode, startIPNode.offset) : __createInsertionPoint(ip.top, startIPnode.parentNode, startIPNode.offset);
			else
				startip = handleCollapsedWhitespace ? documentCreateInsertionPoint(ip.top, startIPNode, 0) : __createInsertionPoint(ip.top, startIPNode, 0);
		}
		else
		{
			startip = ip.clone(); // empty!
			// make sure spans around seed element don't cause problems!
			startBoundary = seed.topInlineAncestor.previousSibling;
		}

		// if non null, make sure start boundary is a not a text node (could be an empty one!)
		while(startBoundary && (startBoundary.nodeType == Node.TEXT_NODE))
			startBoundary = startBoundary.previousSibling;

	}
	// after block element - treat as start
	else
	{
		startBoundary = seed;
		startip = ip.clone();
		var endIPNode = null;
		ipNodeIterator.currentNode = seed;

		// go forward while not at top and in same line
		while(!ipNodeIterator.nextNode() && ipNodeIterator.currentNode)
		{
			endIPNode = ipNodeIterator.currentNode;
			if(endIPNode.nodeType == Node.ELEMENT_NODE)
				elementCount++;
		}

		// at least one node is within the line
		if(endIPNode)
		{
			endBoundary = endIPNode.topInlineAncestor.nextSibling;

			if(endIPNode.nodeType == Node.ELEMENT_NODE)
				endip = handleCollapsedWhitespace ? documentCreateInsertionPoint(ip.top, endIPNode.parentNode, (endIPNode.offset+1)) : __createInsertionPoint(ip.top, endIPnode.parentNode, (endIPNode.offset+1));
			else
				endip = handleCollapsedWhitespace ? documentCreateInsertionPoint(ip.top, endIPNode, endIPNode.nodeValue.length) : __createInsertionPoint(ip.top, endIPNode, endIPNode.nodeValue.length);			
		}
		else // empty
		{
			endip = ip.clone(); // empty!
			// make sure spans around seed element don't cause problems!
			endBoundary = seed.topInlineAncestor.nextSibling;
		}

		// if non null, make sure end boundary is a not a text node (could be an empty one!)
		while(endBoundary && (endBoundary.nodeType == Node.TEXT_NODE))
			endBoundary = endBoundary.nextSibling;
	}

	return new BoundedLine(ip.top, startip, endip, startBoundary, endBoundary, elementCount);
}

/**
 * Non block ip (text node or nil element or after nil element) 
 *
 * - case of after block level element but before inline element: this is treated as a non block ip
 * - these lines are never empty: have at least got "seed" as contents

 MAX - buggy - br within paragraph not detected.
 End not beeing set correctly.
 */
function __createCSSLineFromNonBlockIP(ip, handleCollapsedWhitespace)
{
	var seed = ip.ipReferencedNode;
	var elementCount = (seed.nodeType == Node.ELEMENT_NODE) ? 1 : 0;
	var ipNodeIterator = new __IPNodeIterator(ip.top);

	// first get end ip node - could stay being seed!
	var endIPNode = seed;
	ipNodeIterator.currentNode = seed;
	while(!ipNodeIterator.nextNode() && ipNodeIterator.currentNode)
	{
		endIPNode = ipNodeIterator.currentNode;
		if(endIPNode.nodeType == Node.ELEMENT_NODE)
			elementCount++;
	}

	// get start ip node - could stay being seed!
	var startIPNode = seed;
	ipNodeIterator.currentNode = seed;
	while(!ipNodeIterator.previousNode() && ipNodeIterator.currentNode)
	{
		startIPNode = ipNodeIterator.currentNode;
		if(startIPNode.nodeType == Node.ELEMENT_NODE)
			elementCount++;
	}

	var startNode = startIPNode.topInlineAncestor;
	var startBoundary = startNode.previousSibling;
	// if start boundary then previous could just be blank text node or other clutter - skip it
	// MAX: Added test for comment node, such that comments will be included. 
	// (Added additional ELEMENT_NODE test because .computedStyleValue for a comment node throws a wobbly! Not sure if this fix covers all bases.)
	while(startBoundary &&
	      ((startBoundary.nodeType == Node.TEXT_NODE) ||
	       (startBoundary.nodeType == Node.COMMENT_NODE) ||
	       (startBoundary.nodeType == Node.ELEMENT_NODE && startBoundary.computedStyleValue("display")=="inline")) ) 
	{
		startBoundary = startBoundary.previousSibling;
	}

	var startIPOffset = 0;
	if(startIPNode.nodeType == Node.ELEMENT_NODE)
	{
		startIPOffset = startIPNode.offset;
		startIPNode = startIPNode.parentNode;
	}

	var endNode = endIPNode.topInlineAncestor;
	var endBoundary = endNode.nextSibling;
	// if end boundary is non null then next could just be blank text node or other clutter - skip it
	while(endBoundary &&
	      ((endBoundary.nodeType == Node.TEXT_NODE) ||
			   (endBoundary.computedStyleValue("display") == "inline")) )
				 //MAX - do these tests need improving???
	{
		endBoundary = endBoundary.nextSibling;
	}

	var endIPOffset;
	if(endIPNode.nodeType == Node.ELEMENT_NODE)
	{
		endIPOffset = endIPNode.offset+1;
		endIPNode = endIPNode.parentNode;
	}
	else
		endIPOffset = endIPNode.nodeValue.length;

	var startip;
	var endip;
	if(handleCollapsedWhitespace)
	{
		startip = documentCreateInsertionPoint(ip.top, startIPNode, startIPOffset);
		endip = documentCreateInsertionPoint(ip.top, endIPNode, endIPOffset);
	}
	else // don't handle whitespace so make IP's at exact points we start and end with (only used internally)
	{
		startip = __createInsertionPoint(ip.top, startIPNode, startIPOffset);
		endip = __createInsertionPoint(ip.top, endIPNode, endIPOffset);
	}	

	// if there is either a start or end boundary then this is a bounded line
	if(endBoundary || startBoundary)
		return new BoundedLine(ip.top, startip, endip, startBoundary, endBoundary, elementCount);
	return new ContainedLine(ip.top, startip, endip, startNode.parentNode, elementCount);
}

/**
 * Base class
 */
function CSSLine(){}

CSSLine.prototype.init = function(top, startip, endip, lineRange, elementCount, lineParent)
{
	this.__top = top;
	this.__firstInsertionPoint = startip; // only put into subclasses
	this.__lastInsertionPoint = endip;
	this.__lineRange = lineRange;
	this.__elementCount = elementCount;
	this.__lineParent = lineParent;
}

/**
1.
@final
*/
CSSLine.BOUNDED_LINE = 1;
/**
2.
@final
*/
CSSLine.CONTAINED_LINE = 2;

/**
Gets the line type (CONTAINED_LINE or BOUNDED_LINE).

@return lineType
@type number
*/
CSSLine.prototype.__defineGetter__(
	"lineType",
	function() {
		if(this instanceof ContainedLine)
			return CSSLine.CONTAINED_LINE;
		return CSSLine.BOUNDED_LINE;
	}
);

CSSLine.prototype.__defineGetter__("top",	function(){return this.__top});
CSSLine.prototype.__defineGetter__("lineParent", function(){return this.__lineParent});

/**
Gets the line contents of <code>lineParent</code> from the first IP to the last IP.
Return DocumentFragment of line's contents: if empty then empty document fragment is returned.
Note that this will not include any collapsed whitespace that may appear at the beginning or end of a line.

@return lineContents
@type DocumentFragment
*/ 
CSSLine.prototype.__defineGetter__(
	'lineContents',
	function()
	{
/*
		// MAX: Buggy, doesn't get the span if for example <p><span>text</span></p>, only gets the "text" node.
		var contentsRange = document.createRange();
		contentsRange.selectNodeContents(this.lineParent);
		contentsRange.setStart(this.firstInsertionPoint.ipNode, this.firstInsertionPoint.ipOffset);
		contentsRange.setEnd(this.lastInsertionPoint.ipNode, this.lastInsertionPoint.ipOffset);
*/
		var contentsRange = this.lineRange;
		var contents = contentsRange.cloneContents();
		contentsRange.detach();
		return contents;
	}
);

/**
Returns a Range equivalent to the extent of the line: this does contain collapsed whitespace.

@return lineRange
@type Range
*/ 
CSSLine.prototype.__defineGetter__(
	'lineRange',
	function()
	{
		return(this.__lineRange.cloneRange());
	}
);

CSSLine.prototype.equivalent = function(otherLine)
{
	if(this.firstInsertionPoint.equivalent(otherLine.firstInsertionPoint))
		return true;

	return false;
}

/**
Returns the line as a string.

@return line text.
@type String
*/
CSSLine.prototype.toString = function()
{
	return this.__lineRange.toString();
}

/**
Gets the line length (text length + <code>elementCount</code>).

@return length
@type Integer
*/
CSSLine.prototype.__defineGetter__(
	"length",
	function()
	{
		return(this.toString().length + this.__elementCount);
	}
);

/**
Return the insertion point at a particular offset within a line.
*/
CSSLine.prototype.insertionPointAt = function(lineOffset)
{
	var ip = this.firstInsertionPoint;

	var count = lineOffset;

	while(count)
	{	
		var thisip = ip.clone();
		var result = ip.forwardOne();
		if(result != InsertionPoint.SAME_LINE)
		{
			if(result == InsertionPoint.CROSSED_BLOCK)
				ip.set(thisip);
			break;
		}
		count--;
	}

	ip.__markedOffset = lineOffset;
	
	return ip;
}

/**
Is this an empty line?
A line is empty if the first and last ip's are the equivalent: it may only have
a token or can actually be contentless but one of its boundaries is selectable.
*/
CSSLine.prototype.__defineGetter__(
	"emptyLine",
	function()
	{
		return(this.firstInsertionPoint.equivalent(this.lastInsertionPoint));
	}
);

/**
Is this a Token Line?
Token line - different from an empty line.
A token line is always an empty line but an empty line may not be a token line.
*/
CSSLine.prototype.__defineGetter__(
	"tokenLine",
	function()
	{
		if(this.__elementCount > 0)
			return false;

		// See if any non whitespace/non nbsp characters - if none then this is a token line
		var lineString = this.__lineRange.toString();

		if(/\S+/.test(lineString)) // any non white space visible characters
			return false;

		// of remaining: must have one and only one nbsp
		var matches = lineString.match(/([\u00A0])/g);
//		var matches = lineString.match(/([STRING_TOKEN])/g);
		if(!matches || (matches.length > 1))
			return false;
		return true;
	}
);

/**
 * Base class sets default of "topLine" to false.
 */
CSSLine.prototype.__defineGetter__(
	"topLine",
	function()
	{
		return false;
	}
);

CSSLine.prototype.__defineGetter__(
	"firstInsertionPoint",
	function()
	{
		return this.__firstInsertionPoint.clone();
	}
);

CSSLine.prototype.__defineGetter__(
	"lastInsertionPoint",
	function()
	{
		return this.__lastInsertionPoint.clone();
	}
);

CSSLine.prototype.__defineGetter__(
	"previousLine",
	function()
	{
		var firstIP = this.firstInsertionPoint;
		var prevLineLastIP = firstIP.clone();
		var result = prevLineLastIP.backOne();
		if(result == InsertionPoint.AT_TOP)
			return null;
		var pline = documentCreateCSSLine(prevLineLastIP);
		return pline;
	}
);

CSSLine.prototype.__defineGetter__(
	"nextLine",
	function()
	{
		var lastIP = this.lastInsertionPoint;
		var nextLineFirstIP = lastIP.clone();
		var result = nextLineFirstIP.forwardOne();
		if(result == InsertionPoint.AT_TOP)
			return null;
		var nextLine = documentCreateCSSLine(nextLineFirstIP);
		return nextLine;	
	}
);

/**
Gets the ancestor Element node with a CSS display value of "list-item" for this line.
 
@return list-item element ancestor, or null if no list-item ancestor.
*/
CSSLine.prototype.__defineGetter__(
	"listItemAncestor",
	function()
	{
		var n = this.__lineParent;
		while(n != document.documentElement) {
			if(n.nodeType == Node.ELEMENT_NODE) {
				if(n.computedStyleValue("display") == "list-item") {
					return n;
				}
			}
			n = n.parentNode;				
		}
		return null;
	}
);

/**
 * @returns the table cell element that contains this line or that is an ancestor of this line's container or null
 * if there is no such element.
 */
CSSLine.prototype.__defineGetter__(
	"tableCellAncestor",
	function()
	{
		var nodeToTest = this.__lineParent;
		while(nodeToTest != this.__top)
		{
			if(nodeToTest.computedStyleValue("display") == "table-cell")
				return nodeToTest;
			nodeToTest = nodeToTest.parentNode;				
		}
		return null;
	}
);

CSSLine.prototype.setTopToTableCellAncestor = function()
{
	var tca = this.tableCellAncestor;
	if(tca)
	{
		this.__top = tca;
		this.__firstInsertionPoint.resetTop(tca);
		this.__lastInsertionPoint.resetTop(tca);
	}
}

CSSLine.prototype.containsNode = function(nodeToTest)
{
	var result = false;
	var testRange = document.createRange();
	testRange.selectNodeContents(nodeToTest);
	if(this.__lineRange.containsRange(testRange))
		result=true;
	testRange.detach();
	return result;
}

CSSLine.prototype.normalizeTextNodes = function()
{
	this.lineParent.normalize();
	this.__recalculateLastInsertionPoint();

	return this;
}

/**
XHTML specific : Normalizes "span" elements.
This looks to be incomplete.
POST05: implement properly to stick within line: - ie/ do specialized version for bounded line
*/
CSSLine.prototype.normalizeInlineElements = function()
{
	// then normalize the inline elements
	var normalizeRange = document.createRange();
	normalizeRange.selectNode(this.lineParent);
	normalizeRange.normalizeElements("span");
	this.__recalculateLastInsertionPoint();

	return this;
}

CSSLine.prototype.__recalculateLastInsertionPoint = function()
{
	var referencedNode = this.firstInsertionPoint.ipReferencedNode;

	var ipNodeIterator = new __IPNodeIterator(this.lineParent);
	var endIPNode = ((referencedNode.nodeType == Node.TEXT_NODE) || (referencedNode.computedStyleValue("display") == "inline")) ? referencedNode :  null;
	ipNodeIterator.currentNode = referencedNode;
	while(!ipNodeIterator.nextNode() && ipNodeIterator.currentNode)
		endIPNode = ipNodeIterator.currentNode;
	var endIPOffset;
	if(!endIPNode)
		this.__lastInsertionPoint = this.__firstInsertionPoint.clone();
	else
	{ 
		if(endIPNode.nodeType == Node.ELEMENT_NODE)
		{
			endIPOffset = endIPNode.offset+1;
			endIPNode = endIPNode.parentNode;
		}
		else
			endIPOffset = endIPNode.nodeValue.length;

		this.__lastInsertionPoint = documentCreateInsertionPoint(this.top, endIPNode, endIPOffset);
	}

	return this.__lastInsertionPoint;
}

/*************************************** ContainedLine *************************************************/

ContainedLine.prototype = new CSSLine();
ContainedLine.prototype.constructor = ContainedLine;
ContainedLine.superclass = CSSLine.prototype;
/**
eDOM: Create a Contained Line.
This is a CSSLine with the additional property <code>container</code>.
<code>container</code> is the same as <code>lineParent</code>.

@constructor
*/
function ContainedLine(top, startip, endip, container, elementCount)
{
	var lineRange = document.createRange();
	lineRange.selectNodeContents(container);
	ContainedLine.superclass.init.call(this, top, startip, endip, lineRange, elementCount, container);
	this.__container = container;
}
/**
eDOM: Gets the container (element node) for this Contained Line.
@return container
@type Node
*/
ContainedLine.prototype.__defineGetter__(
	"container",
	function(){return this.__container;	}
);

/**
Gets the highest exclusive container/parent of this line: this may be "top".
<p>Note: topMostContainer can not be above "top".

<p>
E.g. Consider these lists, where ul is a descendent of "top".<br />
 [ul][li]text[/li][li]more[/li][/ul]<br />
for "text" tMC is the container li.<br />
<br />
 [ul][li]text[/li][/ul]<br />
for "text" tMC is now ul, because li is an exclusive (only) child of ul.
</p>

@return topMostContainer
@type Element
*/
ContainedLine.prototype.__defineGetter__(
	"topMostContainer",
	function()
	{
		var topMostContainer = this.__container;
		var parentNode = topMostContainer.parentNode;
		while( (topMostContainer != this.__top) && (oneSignificantChild(parentNode)) ) {
			topMostContainer = parentNode;
			parentNode = parentNode.parentNode;
		}
		return topMostContainer;
	}
);

// Significant means non empty text nodes or elements
function oneSignificantChild(element)
{
	if(!element)
		return false;

	var c = 0;
	var l = element.childNodes.length;
	for(var i=0; i<l; i++) {
		var n = element.childNodes[i];
		if( (n.nodeType == Node.TEXT_NODE) && (__NodeFilter.nonEmptyText(n)==NodeFilter.FILTER_ACCEPT) ) {
			c++;
		}else if( n.nodeType == Node.ELEMENT_NODE ) {
			c++;
		}
		if ( c > 1 ) break;
	}
	return (c==1);
}

/**
eDOM: Is this the top line? 

@return topLine true if <code>container</code> or <code>topMostContainer</code> is <code>top</code>.
@type Boolean
*/
ContainedLine.prototype.__defineGetter__(
	"topLine",
	function() {
		if((this.__container == this.__top) || (this.topMostContainer == this.__top))
			return true;
		return false;
	}
);

/**
eDOM: Is this line embedded within multiple exclusive containers?

@return true if this.container != this.topMostContainer
@type Boolean
*/
ContainedLine.prototype.__defineGetter__(
	"embeddedLine",
	function()
	{
		if(this.topMostContainer != this.container)
			return true;
		return false;
	}
);

/**
0 : Container is block.
@final
*/
ContainedLine.BLOCK = 0;
/**
1 : Container is list-item.
@final
*/
ContainedLine.LIST_ITEM = 1;
/**
2 : container is table-cell.
@final
*/
ContainedLine.TABLE_CELL = 2;
/**
3 : container is top - overrides importance of any other container type.
@final
*/
ContainedLine.TOP = 3;
/**
-1 : Error?
@final
*/
ContainedLine.UNKNOWN = -1;

/**
eDOM: Gets the Contained Line type.
Returns TOP if the top line, otherwise the CSS "display" value for <code>container</code>.

What about the other possible CSS display values?

@return BLOCK | LIST_ITEM | TABLE_CELL | TOP (container==top) | -1
@type {Integer}
*/
ContainedLine.prototype.__defineGetter__(
	"containedLineType",
	function()
	{
		if(this.__container == this.__top)
			return ContainedLine.TOP;

		var containerDisplay = this.__container.computedStyleValue("display");
		if(containerDisplay == "block")
			return ContainedLine.BLOCK;
		else if(containerDisplay == "list-item")
			return ContainedLine.LIST_ITEM;
		else if(containerDisplay == "table-cell")
			return ContainedLine.TABLE_CELL;
		return -1; //There are other valid CSS "display" types...
	}
);

/**
eDOM: Replaces the container element with newContainer, or wraps the contents of
container with newContainer. If container is "top" it is NEVER replaced,
the contents are always wrapped inside newContainer.

@param {Element} newContainer
@param {Boolean} replace true to replace container with newContainer, false to wrap contents with newContainer.
*/
ContainedLine.prototype.setContainer = function(newContainer, replace)
{
	if(this.__container == this.__top)
		replace = false;

	if(replace) {
	// replace current container
		newContainer = this.__container.parentNode.replaceChildOnlyNS(this.__container, newContainer.namespaceURI, newContainer.nodeName);
	} else {
	// usually for table-cell, at top etc
		this.__container.appendChild(newContainer);
		newContainer.appendChild(this.__lineRange.extractContents());
		newContainer.normalize();
	}

	var linenow = documentCreateCSSLine(newContainer.firstInsertionPoint(this.top));
	linenow.forceLineBreaksBeforeAfter();
	return linenow;
}

/**
eDOM: Adds a newline (linefeed) before and/or after the line as required.

POST05: 
- merge with "normalizeLine" to put in invisible whitespace only where formatting is needed.
For now, can't do as selectRange restoration depends not on line offsets but absolute offsets.
*/
ContainedLine.prototype.forceLineBreaksBeforeAfter = function()
{
	if(this.__container == this.__top)
		return;

	if(this.containedLineType == ContainedLine.TABLE_CELL)
		return;

	// insert a newline before the line's container if there isn't one there already
	// - no prev sib
	// - prev sib is element
	// - prev sib is text but doesn't end in a newline!
	var sibling = this.container.previousSibling;
	if(!sibling || (sibling.nodeType == Node.ELEMENT_NODE) || !(/\n$/.test(sibling.nodeValue))) {
		this.container.parentNode.insertBefore(eDOM.newLine(), this.container);
		this.container.parentNode.normalize();
	}

	sibling = this.container.nextSibling;
	if(!sibling || (sibling.nodeType == Node.ELEMENT_NODE) || !(/^\n/.test(sibling.nodeValue))) {
		this.container.parentNode.insertAfter(eDOM.newLine(), this.container);
		this.container.parentNode.normalize();
	}
}

/*
eDOM: Removes the container from this Contained Line.
Removing a container of a line may make it a BoundedLine or may just remove an intermediate container
and so reset the ContainedLine's container. Equivalent to "removeStartBoundary" for BoundedLine.
Note that removing the container of an empty line will delete the line: in this case, this method returns null.
Doesn't do anything if the container it "top".
*/
ContainedLine.prototype.removeContainer = function()
{
	if(this.__container == this.__top)
		return this; // POST05: throw exception

	if(this.emptyLine) {
		this.deleteLine();
		return null;
	}

	var lineSeedIP = this.firstInsertionPoint;

	var result = lineSeedIP.backOne();

	if(result == InsertionPoint.AT_TOP)
		lineSeedIP = null;

	var containerParent = this.container.parentNode;
	containerParent.insertBefore(this.lineContents, this.container);
	containerParent.removeChild(this.container);
	//MAX: must normalize any text nodes when container is removed.
	//E.g. removing container can typically expose newlines that must be merged.
	containerParent.normalize();
	
	if(lineSeedIP)
		lineSeedIP.forwardOne();
	else
		lineSeedIP = this.top.firstInsertionPoint(this.top);

	return documentCreateCSSLine(lineSeedIP);
}

/**
 * This method does three things for a non-pre line
 * - removes collapsed whitespace at the beginning of the line
 * - removes collapsed whitespace at the end of the line
 * - POST05: removes collapsed whitespace within the line
 * - POST05: work on empty lines: see if token and work from there
 */
ContainedLine.prototype.normalizeWhitespace = function()
{
	if(this.container.computedStyleValue("white-space") == "pre")
		return;

	// for now: don't normalize empty lines
	if(this.emptyLine)
		return;

	// first normalize end: before start as if start is in same text node, it may make lastInsertionPoint invalid
	var range = document.createRange();
	range.selectNode(this.container.lastChild);
	range.setStart(this.__lastInsertionPoint.ipNode, this.__lastInsertionPoint.ipOffset);
	range.deleteContents();

	// then normalize start
	range.selectNode(this.container.firstChild);
	range.setEnd(this.__firstInsertionPoint.ipNode, this.__firstInsertionPoint.ipOffset);
	range.deleteContents();
	this.__firstInsertionPoint.ipOffset = 0;

	// set end after deletion of start - covers complication when start and end are in same node
	if(this.__lastInsertionPoint.ipNode.nodeType == Node.ELEMENT_NODE)
		this.__lastInsertionPoint.ipOffset = this.container.childNodes.length;
	else
		this.__lastInsertionPoint.ipOffset = this.__lastInsertionPoint.ipNode.nodeValue.length;

	// Geiko bug?: leaves empty text nodes after normalize
	if((this.container.lastChild.nodeType == Node.TEXT_NODE) && (this.container.lastChild.nodeValue.length == 0))
		this.container.removeChild(this.container.lastChild);

	// reset Range
	this.__lineRange.selectNodeContents(this.container);

	return;
}

/**
eDOM: Deletes this Contained Line, including exclusive parent's of container.
[Bounded line has "deleteContents" but nothing else.]

TODO: only delete contents of table cell's? [perhaps table cells turn into "top" automatically!]
*/
ContainedLine.prototype.deleteLine = function()
{
	var rangeToDelete = document.createRange();
	var topMostContainer = this.topMostContainer;
	if(topMostContainer == this.__top)
		rangeToDelete.selectNodeContents(topMostContainer);
	else		
		rangeToDelete.selectNode(topMostContainer);
	rangeToDelete.deleteContents();
	rangeToDelete.detach();
}

/**
eDOM: Deletes the contents of this Contained Line and replaces it with an editing token.
i.e. the line becomes a token line.
*/
ContainedLine.prototype.deleteContents = function()
{
	// empty out contents
	this.__lineRange.deleteContents();
	var tokenTextNode = eDOM.token();
	this.__container.appendChild(tokenTextNode); 
	return documentCreateCSSLine(__createInsertionPoint(this.__top, tokenTextNode, 0));
}

/**
eDOM: Appends contents to this Contained Line.
If this line is a Token Line the Token is deleted.
If a text node is appended to an existing text node the line container is normalized.

POST05: rebuild line - is still contained.

@param {Node} contents
@return ip new insertion point.
*/
ContainedLine.prototype.appendContent = function(contents)
{
	// first normalize whitespace in line
	this.normalizeWhitespace();

	if(this.emptyLine) {
	//If empty (i.e. Token) line.
		this.__lineRange.deleteContents();		//Delete the token.
		this.container.appendChild(contents);	//Add the new content
//		var ip = __createInsertionPoint(this.__top, this.container.firstChild, 0);
		var ip = this.container.firstInsertionPoint( this.__top ); //Recalculate the new insertion point. (Start of line)
		return ip;
	}

	// then append contents that are assumed to be only text or inline elements and assumed to be normalized for whitespace
	this.container.appendChild(contents);
	var c = this.container.childNodes;
	var n = c.length;
	if ( n >= 2 && c[n-1].nodeType == Node.TEXT_NODE && c[n-2].nodeType == Node.TEXT_NODE)
	//If adjacent text nodes then normalize them.
		this.container.normalize();

	// give back "join" point
	return this.lastInsertionPoint.clone();
}

/**
eDOM: Applies CSS Style to the container of this Contained Line. 
 
POST05:
- text-align in Mozilla seems to have some CSS3 support - default is "start".
 Some but not all - supports "start" but not "end". Need to handle cases like this.
 http://www.w3.org/TR/css3-text/#alignment-prop

@param {String} styleName
@param {String} styleValue
*/
ContainedLine.prototype.setStyle = function(styleName, styleValue)
{
	this.__container.setStyle(styleName, styleValue);
}

/**
Add or remove a class.
Allow multiple classes.
Not yet implemented...
*/

/**
eDOM: Delete handler when selection is at the start of a contained line.

*/
ContainedLine.prototype.deleteStructure = function()
{
  var ip, line;
	// if this is in a table cell then treat table cell as top! Can't delete beyond the cell!
	var tba = this.tableCellAncestor;
	if(tba)
		this.setTopToTableCellAncestor();

	// Case 1: if container is top do nothing.
	if(this.__container == this.__top)
		return this.firstInsertionPoint;

	var pline = this.previousLine;
	// Case 2: this is a token line and not only line in top (caught above!)
	if(this.emptyLine)
	{
		var top = this.__top;
		this.deleteLine();
		ip = pline ? pline.lastInsertionPoint: top.firstInsertionPoint(top); //Max: BUG : missing top from function call.
		return ip;		
	}
	
	// Case 3: this in list item: special case handling for it
	var la = this.listItemAncestor;
	if(la)
	{
		// Sub case 1: contained line within a list item
		if(this.containedLineType != ContainedLine.LIST_ITEM)
			return this.removeContainer().firstInsertionPoint;
		
		// Sub case 2: there is no list item ancestor before this one
		if(!pline || !pline.listItemAncestor)
			return this.firstInsertionPoint; // POST05: do outdent?

		// Sub case 3: merge into previous list-item (works for any bounded stuff too!)
		ip = pline.appendContent(this.lineContents);
		this.deleteLine(); // remove line from its topmost container downwards
		return ip;
	}

	// Case 4: (not list-item!) if topmostContainer is top then remove container and return
	if(this.topMostContainer == this.top)
	{
		line = this.removeContainer();
		return line.firstInsertionPoint;
	}	

	// Case 5: no previous line - so we're at the start
	if(!pline)
	{
		line = this.removeContainer(); // this isn't token line - caught above
		return line.firstInsertionPoint;
	}

	// Case 6: table before this line
	var ptca = pline.tableCellAncestor;
	if(ptca && (ptca != this.top)) // rem: if same tca as this then this.top is same!
	{
		line = this.removeContainer(); // this isn't token line - caught above
		return line.firstInsertionPoint;		
	}

	// Case 7: pline is bounded with an end boundary of an empty element or this line's top most container
	if((pline.lineType == CSSLine.BOUNDED_LINE) && pline.endBoundary && 
           ((pline.endBoundary.contentType == Element.EMPTY_CONTENTTYPE) || (pline.endBoundary == this.topMostContainer)))
	{
		line = this.removeContainer(); // this isn't token line - caught above
		return line.firstInsertionPoint;
	}	

	// Case 8: merge contents into previous line and delete this line	
	ip = pline.appendContent(this.lineContents);

	// remove line from its topmost container downwards
	this.deleteLine();
	return ip;
}

/******************************************* bounded line *******************************************/

/**
 * BoundedLine
 *
 * Processed differently depending on whether it has non-null boundaries, whether those boundaries
 * are EMPTY elements or whether those boundaries have selectable children. 
 */
BoundedLine.prototype = new CSSLine();
BoundedLine.prototype.constructor = BoundedLine;
BoundedLine.superclass = CSSLine.prototype;

function BoundedLine(top, startip, endip, startBoundary, endBoundary, elementCount)
{
	this.__startBoundary = startBoundary;
	this.__endBoundary = endBoundary;

	var lineParentNode = (this.startBoundary != null) ? this.startBoundary.parentNode: this.endBoundary.parentNode;
	var lineRange = document.createRange();
	lineRange.selectNodeContents(lineParentNode);
	if(this.startBoundary)
		lineRange.setStartAfter(this.startBoundary)
	else
		lineRange.setStartBefore(lineParentNode.firstChild);
	
	if(this.endBoundary)
		lineRange.setEndBefore(this.endBoundary);
	else
		lineRange.setEndAfter(lineParentNode.lastChild);

	BoundedLine.superclass.init.call(this, top, startip, endip, lineRange, elementCount, lineParentNode);
}

BoundedLine.prototype.toString = function()
{
	return("sb: " + this.startBoundary + "= eb: " + this.endBoundary + "= lineParent: " + this.lineParent + "= contents: " + this.lineRange.toString());
}

/** 
 * Boundary only line: A line has a start boundary if it is bounded and it is not the first line within an element
 *
 * Note: this can be null but if it is then endBoundary is not null
 */
BoundedLine.prototype.__defineGetter__(
	"startBoundary",
	function() 
	{
		return this.__startBoundary;
	}
);

/**
 * Boundary only line: A line has an end boundary if it is bounded and it is not the last line of an element
 *
 * Note: this can be null but if it is then startBoundary cannot be null
 */
BoundedLine.prototype.__defineGetter__(
	"endBoundary",
	function()
	{
		return this.__endBoundary;	
	}
);

/**
Create a ContainedLine out of the contents of this line: this accounts for empty bounded lines.

MAX : Don't forget that you will need to call ContainedLine.normalizeWhitespace after
this otherwise you may be left with newline chars at either end of the line that are
not selectable. Maybe this should be done automatically before this function returns?

MAX: Warning: If iterating through a line list and changing line containers then
startBoundary.parentNode may no longer exist.
*/
BoundedLine.prototype.setContainer = function(lineContainer, replace)
{
	// noop for empty lines
	if(this.emptyLine) //What if it's a Token line, surely you can set the container?
		return this;

	var lineContents = this.__lineRange.extractContents();
	
	if(this.__startBoundary && this.__startBoundary.parentNode)
		this.__startBoundary.parentNode.insertAfter(lineContainer, this.__startBoundary);
	else if(this.__endBoundary && this.__endBoundary.parentNode)
		this.__endBoundary.parentNode.insertBefore(lineContainer, this.__endBoundary);
	else // ...this shouldn't ever happen.
		return this;

	lineContainer.appendChild(lineContents);
	lineContainer.normalize(); //Just in case Bounded Line was not normalized...

	var ip = lineContainer.firstInsertionPoint(this.top);
	var newLine = document.createCSSLine( ip );

	newLine.forceLineBreaksBeforeAfter();
	return newLine;
}

/**
 * This method does three things for a non-pre line
 * - removes collapsed whitespace at the beginning of the line
 * - removes collapsed whitespace at the end of the line
 * - POST05: removes collapsed whitespace within the line
 */
BoundedLine.prototype.normalizeWhitespace = function()
{
	var lineParentNode = this.lineParent;
	if(lineParentNode.computedStyleValue("white-space") == "pre")
		return;

	// for now - don't normalize token lines
	if(this.emptyLine)
		return;

	var range = document.createRange();
	range.selectNodeContents(lineParentNode);
	range.setStart(this.lastInsertionPoint.ipNode, this.lastInsertionPoint.ipOffset);
	if(this.endBoundary)
		range.setEndBefore(this.endBoundary);
	else
		range.setEndAfter(lineParentNode.lastChild);
	range.deleteContents();

	// then normalize start
	range.selectNodeContents(lineParentNode);
	if(this.startBoundary)
		range.setStartAfter(this.startBoundary)
	else
		range.setStartBefore(lineParentNode.firstChild);
	range.setEnd(this.firstInsertionPoint.ipNode, this.firstInsertionPoint.ipOffset);
	range.deleteContents();

	// partially selected text insertion points: start will now be gone
	if(this.__firstInsertionPoint.ipNode.nodeType == Node.TEXT_NODE)
		this.__firstInsertionPoint.ipOffset = 0;

	// recalculate last insertion point
	this.__recalculateLastInsertionPoint();

	// reset Range
	this.__lineRange.selectNodeContents(lineParentNode);
	this.__lineRange.setStart(this.firstInsertionPoint.ipNode, this.firstInsertionPoint.ipOffset);
	this.__lineRange.setEnd(this.lastInsertionPoint.ipNode, this.lastInsertionPoint.ipOffset);
	
	/**
	// set end after deletion of start - covers complication when start and end are in same node
	if(this.__lastInsertionPoint.ipNode.nodeType == Node.ELEMENT_NODE)
		this.__lastInsertionPoint.ipOffset = (this.endBoundary) ? this.endBoundary.offset : lineParentNode.childNodes.length;
	else
		this.__lastInsertionPoint.ipOffset = this.lastInsertionPoint.ipNode.nodeValue.length;
	**/
}

// TMP: should really support BoundedLine.topParentNode (ie/ most exclusive parent for a line ...); support "deleteLine" for
// BoundedLine which means start (if exists) + all contents. Would also nix exclusive parents.
Node.prototype.__topExclusiveParentNode = function(max)
{
	if(this == max)
		return null;

	var topExclusiveParentNode = null;
	var parentToTest = this.parentNode;
	while((topExclusiveParentNode != max) && (oneSignificantChild(parentToTest)))
	{
		topExclusiveParentNode = parentToTest;
		parentToTest = parentToTest.parentNode;
	}
	return topExclusiveParentNode;
}

/**
 * Deleting the start boundary will delete the bounded line itself. This only works for start boundary's with EMPTY_CONTENTTYPE.
 * This is equivalent to ContainedLine's "removeContainer".
 */
BoundedLine.prototype.deleteStartBoundary = function()
{
	if(this.__startBoundary && (this.__startBoundary.contentType == Element.EMPTY_CONTENTTYPE))
	{			
		// set return ip before the boundary to be deleted
		var ip = documentCreateInsertionPoint(this.top, this.__startBoundary.parentNode, this.__startBoundary.offset);

		var result = ip.backOne();
		if(result == InsertionPoint.AT_TOP)
			ip = null;

		// case 1: empty line - if exclusive parent then empty that out: and select within it - otherwise
		// delete start and go to end of previous line
		if(this.emptyLine) 
		{
			var nodeToDelete = this.__startBoundary.__topExclusiveParentNode(this.top);
			if(nodeToDelete) 
			{
				var range = document.createRange();
				range.selectNodeContents(nodeToDelete);
				range.deleteContents();
				return nodeToDelete.firstInsertionPoint(this.top);
			}
		}

		this.__startBoundary.parentNode.removeChild(this.__startBoundary);
		if(ip)
			ip.forwardOne();
		else
			ip = this.top.firstInsertionPoint(this.top);
		return ip;
	}

	return this.firstInsertionPoint;
}

// OTHER ISSUES: not start boundary ie/ if null but <div>|<hr>stuff< ... remove div?

/**
 * DeleteContents: the line becomes empty
 */
BoundedLine.prototype.deleteContents = function()
{
	if(this.emptyLine)
		return this; // it's a noop!

	// empty out contents
	this.__lineRange.deleteContents();
	this.__elementCount = 0;

	// case 1: line starts or ends with empty element - this is enough to make an empty line
	var elementToSelect = (this.startBoundary && (this.startBoundary.contentType == Element.EMPTY_CONTENTTYPE)) ? this.startBoundary : null;
	elementToSelect = (!elementToSelect && this.endBoundary && (this.endBoundary.contentType == Element.EMPTY_CONTENTTYPE)) ? this.endBoundary : elementToSelect;
	if(elementToSelect)
	{
		this.__firstInsertionPoint = __createInsertionPoint(this.top, elementToSelect.parentNode, (elementToSelect.offset+1));
		this.__lastInsertionPoint = this.__firstInsertionPoint.clone();
		return this;
	}
		
	// case 2: empty line needs selectable token
	var tokenTextNode = eDOM.token();
	if(this.startBoundary)
		this.startBoundary.parentNode.insertAfter(tokenTextNode, this.startBoundary);
	else
		this.endBoundary.parentNode.insertBefore(tokenTextNode, this.endBoundary);

	this.__firstInsertionPoint = __createInsertionPoint(this.__top, tokenTextNode, 0);
	this.__lastInsertionPoint = this.__firstInsertionPoint.clone();
	this.__lineRange.selectNodeContents(tokenTextNode);
	return this;
}

/**
 * POST05:
 * - rebuild line (reset constants)
 */
BoundedLine.prototype.appendContent = function(contents)
{
	// first normalize whitespace in line
	this.normalizeWhitespace();

	// nix any token if there is one
	if(this.emptyLine && !this.__lineRange.collapsed)
		this.__lineRange.deleteContents();

	if(this.endBoundary)
		this.endBoundary.parentNode.insertBefore(contents, this.endBoundary);
	else
		this.lineParent.appendChild(contents);

	return this.lastInsertionPoint;
}

BoundedLine.prototype.deleteLineParent = function()
{
	var previp = this.lineParent.firstInsertionPoint;
	var result = previp.backOne();
	if(result != InsertionPoint.AT_TOP)
	{
		this.lineParent.parentNode.removeChildOnly(this.lineParent);
		previp.forwardOne();
	}
	return previp;
}

/**
 * Delete handler when selection is at the start of a bounded line
 *
 * This is driven purely by CSS which works for XHTML but is unlike to work for semantically richer markup
 * 
 * - POST05: handle the case of <div>xx</div><div>|<hr>stuff</div> where lineParent is deleted
 * and content of div is put into previous div or perhaps just the line parent is removed. ie/ use deleteLineParent
 */ 
BoundedLine.prototype.deleteStructure = function()
{
  var ip, pline;
	// Case 1: delete empty element if bounded line starts with an empty element
	if(this.startBoundary && (this.startBoundary.contentType == Element.EMPTY_CONTENTTYPE))
	{	
		// a/c for collapsed whitespace that may be exposed - nix it first
		this.normalizeWhitespace();
		pline = this.previousLine;
		if(pline)
			pline.normalizeWhitespace();
		ip = this.deleteStartBoundary();
		return ip;
	}

	// Case 2: no previous line then noop
	pline = this.previousLine;
	if(!pline)
		return this.firstInsertionPoint;

	// Case 3: bounded line or previous line in table cell but both are not!
	var tba = this.tableCellAncestor;
	var ptba = pline.tableCellAncestor;
	if((tba || ptba) && (ptba != tba))
		return this.firstInsertionPoint;

	// Case 4: if previous line's lineParent is not the same as this line's line parent and line is empty ...

	// Case 4: if previous line is contained then merge contents with previous line and delete the line
	if(pline.lineType == CSSLine.CONTAINED_LINE)
	{
		ip = pline.appendContent(this.lineContents);
		this.deleteContents();
		return ip;
	}

	// Case 5: pline is bounded and has an end boundary that is different than this line's start boundary: special case - noop
	if(pline.endBoundary)
		return this.firstInsertionPoint; // May want instead to delete that boundary or delete lineParent of line

	// Case 6: otherwise merge this's contents into pline and empty this
	ip = pline.appendContent(this.lineContents);
	this.deleteContents();
	return ip;
}

/**
 * Turns the bounded line into a token line and returns the token's text node
 */
BoundedLine.prototype.setToTokenLine = function()
{
	if(!this.emptyLine)
		this.deleteContents();

	// could change to insert NBSP!
	var textNode = document.createTextNode(STRING_NBSP);
	if(this.startBoundary)
		this.startBoundary.parentNode.insertAfter(textNode, this.startBoundary);
	else
		this.endBoundary.parentNode.insertBefore(textNode, this.endBoundary);

	return __createInsertionPoint(this.top, textNode, 0);
}

/**
Creates a Marked CSSRange.

<p>CSSRange is a Range with methods driven by CSS.
 
<p>POST05: 
<ul>
<li>rename to be CSSRange (from CSSTextRange)
<li>for now, just implement as a Range
<li>rework to remove any explicit reference to XHTML (move into eDOM XHTML)
<li>creation should make sure that the range boundaries are insertion points
</ul>

@param {Range} range
@param {Element} top
@return CSSRange
@type Range
*/
Document.prototype.createCSSTextRange = documentCreateCSSTextRange;

/**
@private
*/ 
function documentCreateCSSTextRange(range, top)
{	
	return __createCSSRange(range, top, true);
}
/**
@private
*/ 
function __createCSSRange(range, top, mark)
{
	range.__top = top;
	if(mark)
		range.__markTextBoundaries(true);
	return range;
}

/**
Creates a Marked CSSRange from Insertion Points.

@param {Element} top
@param {InsertionPoint} startip
@param {InsertionPoint} endip
@return CSSRange
@type Range
*/
function __createCSSRangeFromIPs(top, startip, endip)
{
	var range = document.createRange();
	range.selectNode(startip.ipNode);
	range.setStart(startip.ipNode, startip.ipOffset);
	range.setEnd(endip.ipNode, endip.ipOffset);
	range = documentCreateCSSTextRange(range, top);
	return range;
}

/*
MAX: ...Duplicated below, with additions...
Range.prototype.__defineGetter__(
	"firstInsertionPoint",
	function()
	{
		return documentCreateInsertionPoint(this.__top, this.startContainer, this.startOffset);
	}
);
*/
Range.prototype.__defineGetter__(
	"lastInsertionPoint",
	function()
	{
		var ip = documentCreateInsertionPoint(this.__top, this.endContainer, this.endOffset);
		return ip;
	}
);
/**
Gets an array of lines.

@return lines (array)
@type CSSLine
*/
Range.prototype.__defineGetter__(
	"lines",
	function()
	{
		var lines = new Array();
		var startip = this.firstInsertionPoint();
		var nextLine = documentCreateCSSLine(startip);
		do 
		{
			lines.push(nextLine);
			nextLine = nextLine.nextLine;
		}
		while(nextLine && (this.containsInsertionPoint(nextLine.firstInsertionPoint)));

		return lines;
	}
);

Range.prototype.containsInsertionPoint = function(ipToTest)
{
	var testRange = this.cloneRange();
	if(testRange.startContainer.nodeType == Node.TEXT_NODE)
		testRange.setStart(testRange.startContainer.parentNode, testRange.startContainer.offset);
	if(testRange.endContainer.nodeType == Node.TEXT_NODE)
		testRange.setEnd(testRange.endContainer.parentNode, (testRange.endContainer.offset+1));
	
	var nodeRange = document.createRange();
	nodeRange.selectNodeContents(ipToTest.ipReferencedNode);

	if(!testRange.containsRange(nodeRange))
		return false; 
	return true;
}

/**
Gets the first insertion point for this range.

Assumes that Range has a valid insertion point???

#return insertion point or null.
*/
Range.prototype.firstInsertionPoint = function() {
	var ip;
	if(this.startContainer.nodeType == Node.TEXT_NODE) {
	//If TEXT_NODE
		//Then the first IP is just the offset in the startContainer.
		ip = documentCreateInsertionPoint(this.top, this.startContainer, this.startOffset);
		return ip;
	}
	else {
	//Else ELEMENT_NODE ... i.e. offset to before/after a child node.
		var ipni, seed;
		if( this.startOffset < this.startContainer.childNodes.length ) {
		//If startOffset is not pointing after the last child;
			seed = this.startContainer.childNodes[this.startOffset];
			if(seed.nodeType == Node.TEXT_NODE) {
			//If TEXT_NODE
				ip = documentCreateInsertionPoint(this.top, seed, 0);
				return ip;
			}
			//Else if Empty ELEMENT_NODE 
			if(seed.contentType == Element.EMPTY_CONTENTTYPE) {
				ip = documentCreateInsertionPoint(this.top, this.startContainer, this.startOffset);
				return ip;;
			}
			//Else if Element, but not empty content-type
			ipni = new __IPNodeIterator(this.startContainer);
			ipni.currentNode = seed;
			ipni.nextNode();
			seed = ipni.currentNode;
			if(!seed) {
				ipni.previousNode();
				seed = ipni.nextNode();
			}
			if(!seed)
				return null;
			ip = documentCreateInsertionPoint(this.top, seed.parentNode, seed.offset);
			return ip;
		}
		else {
		//Else startOffset points after the last child;
			//MAX : This works based on the last child always being an Element node.
			// note: __forwardOne and __backOne always return a Text Node.
			// If other IP creation functions point AFTER a text node this would need to reset to last IP of text node.
			ipni = new __IPNodeIterator(this.startContainer);
			seed = ipni.setToEnd();				
			if(!seed)
				return null;
			ip = documentCreateInsertionPoint(this.top, seed.parentNode, seed.offset);
			return ip;
		}
	}
}


/**
Collapses this range to the point specified by "ip".
 
<p>Note: this can cause problems if you are trying to set the cursor to the point
after an inline empty element.

@param {InsertionPoint} ip
*/
Range.prototype.selectInsertionPoint = function(ip)
{
	this.selectNode(ip.ipNode);
	this.setStart(ip.ipNode, ip.ipOffset);
	this.collapse(true);
}

/**
 * POST05: rename to "topCommonAncestorNode"?
 *
 * @return topmost common ancestor for Range
 */
Range.prototype.__defineGetter__(
	"top",
	function() {return this.__top;}
);

function parentBelow(ancestor, element)
{
	var parentNode = element;
	var parentBelow;
	do
	{
		parentBelow = parentNode;	
		parentNode = parentNode.parentNode;
	}
	while(parentNode != ancestor);

	return parentBelow;
}

/**
 * Moves Range boundaries up to include exclusive parents ie/ parentNodes that only contain the selected Range.
 *
 * POST05: make sure InsertionPoint calculation for a Range a/cs for such higher boundaries
 */
Range.prototype.includeExclusiveParents = function()
{	
	var startip = this.firstInsertionPoint();
	var startLine = startip.line;
	var prevResult = InsertionPoint.SAME_LINE; // default
	var commonAncestorRange;
	// check if already at start of line unless have bounded line with start boundary
	if(!((startLine.lineType == CSSLine.BOUNDED_LINE) && startLine.startBoundary))
	{
		var previp = startip.clone();
		prevResult = previp.backOne();

		if(prevResult == InsertionPoint.AT_TOP) {
			this.setStart(this.top, 0);	
		}
		else if(prevResult == InsertionPoint.CROSSED_BLOCK) {	
			commonAncestorRange = document.createRange();
			commonAncestorRange.setStart(previp.ipNode, previp.ipOffset);
			commonAncestorRange.setEnd(startip.ipNode, startip.ipOffset);
			var startContainer = commonAncestorRange.commonAncestorContainer;
			var startOffset = parentBelow(startContainer, startip.ipNode).offset;
			this.setStart(startContainer, startOffset);
		}
		// SAME_LINE - see if different parent
		else if(previp.ipNode != startip.ipNode) {
			this.setStart(previp.ipNode, (previp.ipOffset+1));
		}
	}

	var endip = this.lastInsertionPoint;
	var endLine = endip.line;
	var nextResult = InsertionPoint.SAME_LINE; // default
	// check if already at start of line unless have bounded line with start boundary
	if(!((endLine.lineType == CSSLine.BOUNDED_LINE) && endLine.endBoundary))
	{
		var nextip = endip.clone();
		nextResult = nextip.forwardOne();

		if(nextResult == InsertionPoint.AT_TOP)
		{
			this.setEnd(this.top, this.top.childNodes.length);
		}
		else if(nextResult == InsertionPoint.CROSSED_BLOCK)
		{
			commonAncestorRange = document.createRange();
			commonAncestorRange.setStart(endip.ipNode, endip.ipOffset);
			commonAncestorRange.setEnd(nextip.ipNode, nextip.ipOffset);
			var endContainer = commonAncestorRange.commonAncestorContainer;
			var endOffset = parentBelow(endContainer, endip.ipNode).offset + 1; // go right beyond the endip.ipNodes parent
			this.setEnd(endContainer, endOffset); 
		}
		//Else SAME_LINE
		else if(nextip.ipNode != endip.ipNode)
		{
		//MAX : nextip.ipOffset is 0 when at the start of a text node, -1 is an invalid offset.
			if (nextip.ipNode.nodeType == Node.TEXT_NODE)
				this.setEnd(nextip.ipNode, (nextip.ipOffset));
			else
				this.setEnd(nextip.ipNode, (nextip.ipOffset-1));
		}
	}

	// maximize covers top?
	return((nextResult == InsertionPoint.AT_TOP) && (prevResult == InsertionPoint.AT_TOP));
}

/**
 * Extract the contents of a Range based on CSS: this is greedy, obeys the top boundary and a/cs for table settings
 */
Range.prototype.extractContentsByCSS = function()
{
	var extractedContents = document.createDocumentFragment();

	var startTableRanges = this.__startTableRanges();
	var endTableRanges = this.__endTableRanges();

	for(var i=0; i<startTableRanges.length; i++)
		extractedContents.appendChild(startTableRanges[i].__simpleExtractContentsByCSS());

	if(!this.collapsed)
		extractedContents.appendChild(this.__simpleExtractContentsByCSS());

	for(i=0; i<endTableRanges.length; i++)
		extractedContents.appendChild(endTableRanges[i].__simpleExtractContentsByCSS());

	// correct default insertion point if there were table cells in line first!
	if(startTableRanges.length)
		this.selectInsertionPoint(startTableRanges[0].firstInsertionPoint());
	
	return extractedContents;
}

/**
 * Don't account for table cells
 *
 * POST05: 
 * - do the ip marker stuff more neatly ie/ see sequence below to account for whitespace. Too complex
 */
Range.prototype.__simpleExtractContentsByCSS = function()
{
	if(this.collapsed) // return empty document fragment
		return document.createDocumentFragment();

	var startip = this.firstInsertionPoint();
	var returnip = startip.clone();
	var partialEndSelection = false;
	var endip = this.lastInsertionPoint;

	if(this.includeExclusiveParents())
	{	
		// special case one: surrounds top!
		this.top.appendChild(document.createTextNode(STRING_NBSP));
		returnip = null;
	}
	// a/c for two things: start of line and partial selection of last line
	else
	{
		var startLine = startip.line;
		var endLine = endip.line;

		// test if crossed block or reference element node: go back one if we do 
		var returnResult = InsertionPoint.SAME_LINE;
		if((this.startContainer.nodeType == Node.ELEMENT_NODE) || (this.startContainer != startip.ipNode))
		{		
			returnResult = returnip.backOne();
			if(returnResult == InsertionPoint.AT_TOP)
				returnip = null;
		}

		// if only partially select end line then must merge unless it is the same as the start line or whole start line is in the range to be deleted
		if(!endip.equivalent(endLine.lastInsertionPoint) && !startLine.equivalent(endLine) && !startip.equivalent(startLine.firstInsertionPoint))
			partialEndSelection = true;
	}

	var extractedContents = this.extractContents(); 

	// note that this will only remove immediate container of a contained line. If a line has multiple levels of 
	// containment then just the first is removed and the line isn't "merged". May change this?
	if(partialEndSelection)
	{
		if(returnip)
		{
			var returnline = returnip.line;
			var returnipoffset = returnip.lineOffset;
			var endlineip = endip.line.deleteStructure();
			returnip = documentCreateCSSLine(endlineip).insertionPointAt(returnipoffset);
		}
		else
			endip.line.deleteStructure();
	}

	// restore
	this.selectInsertionPoint((returnip ? returnip : this.top.firstInsertionPoint(this.top)));

	return extractedContents;
}

Range.prototype.__startTableRanges = function()
{
	var ranges = new Array();
	var startRange;
	while(!this.collapsed)
	{
		var startip = this.firstInsertionPoint();
		var sta = startip.tableCellAncestor;

		if(!sta)
			break;

		var eta = this.lastInsertionPoint.tableCellAncestor;

		if(eta != sta)
		{
			// move this range on beyond this table cell
			var lip = sta.lastInsertionPoint(this.top);
			var nextip = lip.clone();
			nextip.forwardOne();
			this.setStart(nextip.ipNode, nextip.ipOffset);
			startRange = __createCSSRangeFromIPs(sta, startip, lip);
		}
		else // startRange is this Range!
		{	
			startRange = documentCreateCSSTextRange(this.cloneRange(), sta);
			this.collapse(false);
		}

		ranges.push(startRange); // if this range ends up collapsed then all are table ranges ...
	}

	return ranges;
}

Range.prototype.__endTableRanges = function()
{
	var ranges = new Array();

	while(!this.collapsed)
	{
		var endip = this.lastInsertionPoint;
		var eta = endip.tableCellAncestor;

		if(!eta)
			break;

		var sta = this.firstInsertionPoint().tableCellAncestor;

		if(eta != sta)
		{
			// move this range on beyond this table cell
			var fip = eta.firstInsertionPoint(this.top);
			var previp = fip.clone();
			previp.backOne();
			this.setEnd(previp.ipNode, previp.ipOffset);
			endRange = __createCSSRangeFromIPs(eta, fip, endip);
		}
		else // startRange is this Range!
		{
			endRange = documentCreateCSSTextRange(this, eta);
			this.collapse(true);
		}

		ranges.push(endRange); // if this range ends up collapsed then all are table ranges ...
	}

	return ranges;
}

/***************************************** The remainder of eDOM is due to be rewritten POST05 ******************/

/**
eDOM: Gets non-empty Text Nodes for this CSS Range.

Should be able to handle any sort of Range.

@return textNodes
@type Array
*/
Range.prototype.textNodes = function() {

		// normalize safely: sets marker by side effect
		this.normalizeText();

		var currentNode = this.startContainer;
		var goToNextNode = false;

		if( this.startContainer.nodeType==Node.TEXT_NODE ) {
			if (this.startOffset >= this.startContainer.length ) {
				goToNextNode=true;
			} else if (this.startOffset > 0) {
				// Split starting text node at startOffset.
				var newStartContainer = this.startContainer.splitText(this.startOffset);
				this.__restoreTextBoundaries(); // needed to restore endContainer when in same node!
				currentNode = newStartContainer;
			}
		} else if ( this.startContainer.nodeType==Node.ELEMENT_NODE ) {
			if (this.startOffset >= this.startContainer.childNodes.length) {
				currentNode = this.startContainer.childNodes[this.startContainer.childNodes.length-1];
				goToNextNode=true;
			} else {
				currentNode = this.startContainer.childNodes[this.startOffset];
			}
		}
		//Split an end Text Node from 0 to 1 before the end.
		if(this.endContainer.nodeType==Node.TEXT_NODE) {
			if (this.endOffset < this.endContainer.length) {
				this.endContainer.splitText(this.endOffset);
			}
		}

		var textTW = document.createTreeWalker(this.commonAncestorContainer,
							   NodeFilter.SHOW_TEXT,
							   null,
							   false);

		textTW.currentNode = currentNode;
		if ( goToNextNode || (currentNode.nodeType == Node.ELEMENT_NODE) ) {
			textTW.nextNode();
		}
		var testNode = textTW.currentNode;
		var textNodes = new Array();
		while(testNode)
		{
			// must allow empty text to get this far as last node could be empty (offset = 0 before split)
			if(__NodeFilter.nonEmptyText(testNode) == NodeFilter.FILTER_ACCEPT) {
				textNodes.push(testNode);
			}
			if (testNode == this.endContainer) {
				break; // Done!
			}else{
			//Check if testNode is outside range
				var testRange = document.createRange();
				testRange.selectNode( testNode );
				if ( this.compareBoundaryPoints(Range.START_TO_END, testRange) < 1  ) {
					if (textNodes.length>0 && textNodes[textNodes.length-1]==testNode) {
						textNodes.pop();
					}
					break;
				}
			}
			testNode = textTW.nextNode();
		}

		// do one global one at the end?
		this.__restoreTextBoundaries();

		return textNodes;
	}

/**
 * Does one range contain another?
 *
 * @argument compareRange range to check
 * @returns true if one contains the other
 * 
 * POST04: 
 * - try out compareNode of Mozilla. May be no use case for Range in Range - all Node in Range
 * - see if this is called too much (ie/ trace it)
 */
Range.prototype.containsRange = function(compareRange)
{
	// Compare against a maximized version of this range.
	var baseRange = this.cloneRange();
	baseRange.__maximizeContext();			
	
	var val = baseRange.compareBoundaryPoints(Range.START_TO_START, compareRange);

	if(baseRange.compareBoundaryPoints(Range.START_TO_START, compareRange) == 1)
		return false;


	if(baseRange.compareBoundaryPoints(Range.END_TO_END, compareRange) == -1)
		return false;


	return true;	
}

/**
 * recursively combine adjacent editable elements in a Range.
 *
 * POST04:
 * - change to generic: normalizeInlineElements
 * - move out of Range and make this an Element method. Then if necessary do a 
 * Range equivalent that just resets boundaries. Typical use case is to 
 * selectNode and then call this. IMPORTANT
 * - not for text ranges: formalize this - consider general issue of range setting. Caller should expand
 * a range appropriately. This and other methods SHOULD NOT expand it by side effect.
 * - efficiency: only check top most nodes for inclusion in Range: once find included node then don't
 * check their children ... [general pattern for all Range change operations ...]
 */
Range.prototype.normalizeElements = function(elementName)
{
	elementName = elementName.toLowerCase();

	// note: no range markers as range isn't a text range

	// normalize element filter checks two things:
	// - is the element within the Range
	// - does the element have an identical previous sibling (prelude to merge!)
	var range = this;
	var normalizeFilter = function(node)
	{
		var compareRange = document.createRange();
		compareRange.selectNodeContents(node);			
		if(!range.containsRange(compareRange)) {
			return NodeFilter.FILTER_REJECT;
		}
		compareRange.detach();
							
		if(node.nodeName.toLowerCase() == elementName) {
			var previousSibling = node.__editablePreviousSibling;
			if(previousSibling && ("match" in previousSibling) && previousSibling.match(node)) {
//				compareRange = document.createRange();
//				compareRange.selectNodeContents(previousSibling);			
				if(!range.intersectsNode(previousSibling)) { //MAX: Should this be a range intersect rather than complete contain???
					return NodeFilter.FILTER_REJECT;
				} else {
					return NodeFilter.FILTER_ACCEPT;
				}
			}
			return NodeFilter.FILTER_SKIP;
		}	  	
		return NodeFilter.FILTER_SKIP;		
	}

	var namedElTW = document.createTreeWalker(this.commonAncestorContainer,
					   NodeFilter.SHOW_ELEMENT,
					   normalizeFilter,
					   false);

	// recurse through a tree processing the leaf nodes before their parents
	this.normalizeElementsInTree = function(namedElTW)
	{
		// children will all be nodes of the type we want to rename 
		var child = namedElTW.firstChild();

		while(child != null)
		{
			// process the next child before processing its parent!
			this.normalizeElementsInTree(namedElTW);
			namedElTW.currentNode = child; // jump back up!
			var childToMerge = child;
			child = namedElTW.nextSibling(); // moves current on if one more!

			var childToMergeInto = childToMerge.__editablePreviousSibling;
			// delete intermediate empty text nodes
			while(childToMergeInto.nextSibling != childToMerge)
				childToMergeInto.parentNode.removeChild(childToMergeInto.nextSibling);
			var contentsToMerge = document.createRange();
			contentsToMerge.selectNodeContents(childToMerge);
			childToMergeInto.appendChild(contentsToMerge.extractContents());
			childToMergeInto.normalize();
			// now delete the merged child
			childToMerge.parentNode.removeChild(childToMerge);	

		}
	}

	// now normalize elements
	this.normalizeElementsInTree(namedElTW);
}

/**
XHTML Specific : Changes all elements in this range with <code>currentName</code> to <code>newName</code>.
This will NOT normalize elements.

MAX : This is an XHTML specific function used to toggle list lines.
It should not be here in eDOM.js

@param {String} currentName
@param {String} newName
@return true if at least one element renamed, otherwise false.
@type boolean 
*/
Range.prototype.renameElements = function(currentName, newName)
{
	// before processing, let's mark text boundaries if the range has them but only if not already set
	this.__markTextBoundaries(false);

// MAX - XML is case sensitive, so shouldn't force lower case here if this is to be generic XML.
//	currentName = currentName.toLowerCase();
//	newName = newName.toLowerCase();

	// - is the element within the Range
	// - does the element have a particular name
	var range = this;
	var renameFilter = function(node)
	{
		var compareRange = document.createRange();

		compareRange.selectNodeContents(node);			

		
		if(!range.containsRange(compareRange))
			return NodeFilter.FILTER_REJECT;
		compareRange.detach();
		if(node.nodeName.toLowerCase() == currentName) 
			return NodeFilter.FILTER_ACCEPT;
		return NodeFilter.FILTER_SKIP;		
	}

	var namedElTW = document.createTreeWalker(this.commonAncestorContainer,
					   NodeFilter.SHOW_ELEMENT,
					   renameFilter,
					   false);

	// recurse through a tree processing the leaf nodes before their parents
	var atLeastOneRenamed = false;
	this.renameElementsInTree = function(namedElTW)
	{
		// children will all be nodes of the type we want to rename 
		var child = namedElTW.firstChild();

		while(child != null)
		{
			// process the next child before processing its parent!
			this.renameElementsInTree(namedElTW);
			namedElTW.currentNode = child; // jump back up!
			var childToRename = child;
			child = namedElTW.nextSibling(); // moves current on if one more!
			childToRename.parentNode.replaceChildOnlyNS(childToRename, childToRename.namespaceURI, newName);
			atLeastOneRenamed = true;
		}
	}

	// now normalize elements (test result!)
	this.renameElementsInTree(namedElTW);
		
	// now restore the text boundaries
	if(atLeastOneRenamed)
	{
		this.__restoreTextBoundaries();
		return true;
	}
	return false;
}

/**
 * Normalize text in a Range without affecting its boundaries.
 */
Range.prototype.normalizeText = function()
{
	this.__markTextBoundaries(false);
	this.commonAncestorContainer.parentElement.normalize();
	this.__restoreTextBoundaries();
}

var keepTxtNodes = null;

/**
 * Does a Range have a style?
 */
Range.prototype.hasStyle = function(styleName, styleValue)
{
	// note that this can split text nodes
	var textNodes = this.textNodes();
	keepTxtNodes = textNodes;

	for(var i=0; i<textNodes.length; i++)
	{
		var textContainer = textNodes[i].parentNode;
		// if any node doesn't have the style then the Range doesn't have it!
		if(textContainer.computedStyleValue(styleName) != styleValue)
		{
			//this.normalizeText();
			return false;	
		}
	}
	// efficiency given that apply comes next - will be combined with apply POST 03
	// this.normalizeText();
	return true;
}

/**
 * @returns all the elements with a particular display value
 *
 * POST04:
 * - don't expand beyond text but skip the text
 */
Range.prototype.elementsWithDisplay = function(displayValue)
{
	// expand range to test to cover start/end of any text nodes
	var range = this.cloneRange();
	if(range.startContainer.nodeType == Node.TEXT_NODE)
		range.setStart(range.startContainer, 0);
	if(range.endContainer.nodeType == Node.TEXT_NODE)
		range.setEnd(range.endContainer, range.endContainer.nodeValue.length);

	// expand range to test to cover start/end of any text nodes
	var displayFilter = function(element)
	{
		if(element.computedStyleValue("display") != displayValue)
			return NodeFilter.FILTER_SKIP;

		var elementRange = document.createRange();
		elementRange.selectNodeContents(element);
		if(!range.containsRange(elementRange))
			return NodeFilter.FILTER_REJECT; 
		return NodeFilter.FILTER_ACCEPT;
	}

	// Tree to allow us to walk and loop through seeds producing unique lines
	var tw = document.createTreeWalker(range.commonAncestorContainer.parentElement,
					   NodeFilter.SHOW_ELEMENT,
					   displayFilter,
					   false);

	var displayElements = new Array();
	var displayElement = tw.firstChild();
	while(displayElement)
	{
		displayElements.push(displayElement);
		displayElement = tw.nextNode();	 
	}
	return displayElements;
} 
/**
Adds start and end ___TextPointers.

POST05: efficiency: make sure not called too often, in particular from read-only operations
POST05: make public or at least have collapse and other methods reset this method!
*/
Range.prototype.__markTextBoundaries = function(replace)
{
	if(!replace && (this.startTextPointer || this.endTextPointer))
		return;

	if(this.startContainer.nodeType == Node.TEXT_NODE) {
		this.startTextPointer = new __TextPointer(this.__top, this.startContainer, this.startOffset);
		if ( this.startTextPointer.textOffset == this.startTextPointer.textNode.nodeValue.length )
			this.startTextPointer.goToStart(false);
	} else {
		if(this.startTextPointer)
			delete this.startTextPointer;
	}

	// optimization possibility: calculate end from start ie/ combine both into a CSSTextRangeMarker
	if(this.endContainer.nodeType == Node.TEXT_NODE) {
		if(this.collapsed) {
			this.endTextPointer = this.startTextPointer;
		} else {
			this.endTextPointer = new __TextPointer(this.__top, this.endContainer, this.endOffset);
			if ( this.endTextPointer.textOffset == this.endTextPointer.textNode.nodeValue.length )
				this.endTextPointer.goToStart(false);
		}
	} else {
		if(this.endTextPointer)
			delete this.endTextPointer;
	}
}

/**
Re-marks existing text pointers.
Use when the original text node and offset haven't changed, but other nodes have
been added or changed.
*/
Range.prototype.__reMarkTextBoundaries = function ()
{
	if (!(("startTextPointer" in this) && ("endTextPointer" in this)))
		return;
	var s = this.startTextPointer;
	var e = this.endTextPointer;
	if ( this.endTextPointer == this.startTextPointer ) {
		this.startTextPointer = new __TextPointer(s.root, s.textNode, s.textOffset, s.__goToStart);
		this.endTextPointer = this.startTextPointer;
	} else {
		this.startTextPointer = new __TextPointer(s.root, s.textNode, s.textOffset, s.__goToStart);
		this.endTextPointer = new __TextPointer(e.root, e.textNode, e.textOffset, e.__goToStart);
	}
}

/*
Restore a Range to the boundaries set by MarkTextBoundaries if they exist. Otherwise leave Range as is.
 
POST04: could change this to always take up text boundaries even where there is no pointer etc. ie/ textBoundaries?
- move to being a utility/impl method of CSSTextRange
- if one or other end fails then revert to one end; one both fail then ?
*/ 
Range.prototype.__restoreTextBoundaries = function()
{
	if(this.startTextPointer)	{
		var startNode = this.startTextPointer.referencedTextNode(); // avoid order error: start after end as being set
		if ( this.endTextPointer )
			this.selectNode(startNode);
// MAX : Selecting the node causes an error when the endTextPointer doesn't exist because it is an element/offset
// combination rather than text-node/offset reference.
		this.setStart(this.startTextPointer.currentTextNode, this.startTextPointer.currentTextOffset);
	}
	if(this.endTextPointer) {	
		this.setEnd(this.endTextPointer.referencedTextNode(), this.endTextPointer.currentTextOffset);		
	}
}

Range.prototype.__clearTextBoundaries = function()
{
	if(this.startTextPointer)
		delete this.startTextPointer;
	if(this.endTextPointer)
		delete this.endTextPointer;
}


Range.prototype.__maximizeContext = function()
{
	var selOffset, parentNode, parentOffset;
	// is start at beginning or end of text node?
	if((this.startContainer.nodeType == Node.TEXT_NODE) &&
	   ((this.startOffset == 0) || (this.startOffset == this.startContainer.nodeValue.length)))
	{
		// at beginning or end? Decides how to index within its context
		selOffset = 1;
		if(this.startOffset == 0)
			selOffset = 0;

		// get in terms of parent
		parentNode = this.startContainer.parentNode;
		parentOffset = this.startContainer.offset;	
		if(this.collapsed)
		{
			this.selectNode(parentNode); // makes sure start is temporarily put after end
			this.setStart(parentNode, parentOffset+selOffset);
			this.collapse(true);
			return;
		}
		this.setStart(parentNode, parentOffset+selOffset);	
	}
	// is end at beginning or end of text node?
	if((this.endContainer.nodeType == Node.TEXT_NODE) &&
	   ((this.endOffset == 0) || (this.endOffset == this.endContainer.nodeValue.length)))
	{
		// at beginning or end? Decides how to index within its context
		selOffset = 1;
		if(this.endOffset == 0)
			selOffset = 0;
		// get in terms of parent
		parentNode = this.endContainer.parentNode;
		parentOffset = this.endContainer.offset;
		this.setEnd(parentNode, parentOffset+selOffset);			
	}
}

/******************************** TextPointer ****************************************/

/**
 * Many methods (appendChild, normalize) that don't remove text from the DOM, do delete
 * a particular TextNode that holds the text at one time or another. They affect 
 * boundaries of Ranges if a Range begins or ends in a TextNode they manipulate. This
 * object, by recording an absolute offset for text, allows the restoration of a text
 * pointer even after specific text nodes disappear.
 *
 * Uses:
 * - restore a Range after expansion or contraction 
 * - restore a Range after its ends are affected by splitting or normalizing or being moved.
 * 
 * Two accessors:
 * - referencedTextNode
 * - referencedOffset
 *
 * POST02:
 * - CSSTextRangeMarker may replace this: calculates end from start and start from a context
 * logical for a Range.
 * - issue: goToStart etc doesn't follow InsertionPoint logic and so treats whitespace as significant
 * - look into using xpointer like syntax for debugging: ie/ id/1/2 etc. Alternatively
 * may implement using xpointers where the child numbers change based on note manipulation
 * - look into dropping skip of non empty text (ie/ orphan white space) because it may
 * be orphaned by applying styles after having been part of a text node. That means that
 * restoration after applying styles may fail. Alternatively, may make routines like "mark"
 * jump to the "real" start or end of a line if already at the effective start or end of a
 * line
 * - may add ability to decide whether to end in one node or begin in a subsequent node. Right
 * now, the default always begins in a new node if it can. 
 * - maybe allow increment and decrement of offset to reflect user controller insertion
 * and deletion of preceding text of a particular size.
 
 @param {Node} root
 @param {Node} text node to point to.
 @param {Integer} offset within text node. 
 @param {Boolean} toStart (optional)
 */ 
function __TextPointer(root, text, offset, toStart)
{
	// public properties of this marker
	this.root = root;
	this.absTextOffset = 0;
	this.textNode = text; // original text node - may go away doing processing
	this.textOffset = offset; // original offset within text
	this.currentTextNode = text;
	this.currentTextOffset = offset;
	this.__goToStart = (arguments.length == 4) ? toStart : true;

	this.textTW = document.createTreeWalker(this.root,
					       NodeFilter.SHOW_TEXT,
					       __NodeFilter.nonEmptyText,
					       false);

	this.absTextOffset = 0;
	this.textTW.currentNode = this.root;
	for(var next = this.textTW.firstChild(); next != null; next = this.textTW.nextNode()) {
		if(next == this.textNode) {
			this.absTextOffset += this.textOffset;
			break;
		}
		this.absTextOffset += next.nodeValue.length;
	}

	/*
   * Now know the absolute offset. If no text deletion then this makes text position
   * recoverable.
   * 
   * The following lets us work back from calculated offset to the current active 
   * text node at any time
   */
	// POST04: should do alert if can't find a TextNode - exception back to caller
	/**
	Recalculates currentTextNode and currentTextOffset. 
	*/
	this.calculateTextNode = function()
	{
		this.textTW.currentNode = this.root;
		var lastTextNode = null; // null means no TextNode is referenced
		var endSoFar = 0;
		var possEndSoFar = 0;

		for(var currentTextNode = this.textTW.firstChild(); currentTextNode != null; currentTextNode = this.textTW.nextNode())
		{
			// at start of node - return with node and offset of 0
			if(this.__goToStart && (this.absTextOffset == possEndSoFar))	
			{
				this.currentTextNode = currentTextNode;
				this.currentTextOffset = 0;
				return;
			}

			possEndSoFar += currentTextNode.nodeValue.length;

			// if now within a node then return that node and the offset within it
			if((this.absTextOffset < possEndSoFar) || ((!this.__goToStart) && (this.absTextOffset == possEndSoFar)))
			{
				this.currentTextNode = currentTextNode;
				this.currentTextOffset = this.absTextOffset - endSoFar;
				return;
			}		

			endSoFar = possEndSoFar;
			lastTextNode = currentTextNode;
		}
		// don't like this: get here if goToStart is true but "start" is outside editable area
		this.currentTextNode = lastTextNode;
		this.currentTextOffset = lastTextNode.nodeValue.length;
		return;
	}
}
/*
The pointer could lead to the end of a text node or the start of the following node.
This setting decides which way to go.

goToStart
true : after the last character of a text node the pointer will be set
to the first character of the next text node.
false : after the last character of a text node the pointer will be set
to after the last character.

POST04: remove need for this or build into pointer creation.

@param value
@type Boolean 
*/
__TextPointer.prototype.goToStart = function (value) {
	this.__goToStart = value;
}
/* Getters and Setters not working for some reason (browser bug?) */
/*
__TextPointer.prototype.__defineSetter__(
	"goToStart",
	function(value) {this.__goToStart = value;}
);
*/
/**
Return the offset (currentTextOffset) within the text node currently referenced by the pointer.
The currentTextOffset is not recalculated.

@return currentTextOffset
@type integer
*/ 
__TextPointer.prototype.referencedOffset = function() {
	return this.currentTextOffset;
}
/*
__TextPointer.prototype.__defineGetter__(
	"referencedOffset",
	function() {
		return this.currentTextOffset;
	}
);
*/
/**
Return the text node (currentTextNode) currently referenced by the pointer.
Note that the pointer is recalculated before being returned. 
POST04:
- separate calculation from accessors

@return referencedTextNode
@type Node
*/
__TextPointer.prototype.referencedTextNode = function() {
	this.calculateTextNode();
	return this.currentTextNode;
}
/*
__TextPointer.prototype.__defineGetter__(
	"referencedTextNode",
	function() {
		this.calculateTextNode();
		return this.currentTextNode;
	}
);
*/

/*
 * Pretty up toString
 */ 
__TextPointer.prototype.toString = function()
{
	return("TextPointer: " + this.referencedTextNode().nodeValue + ":" + this.referencedOffset());
}

/******************************************************************************/
/**
 __NodeFilter provides centralized access to frequently used node filters. They can be 
applied directly, within filter enabled methods and in tree traversal.

 POST05:
 - will go as Range methods are tightened up 

@constructor
*/
function __NodeFilter() {} // Class to hold the filters

/*
Is this node empty?
<p>Need to filter out empty text nodes, those only in the DOM to format raw XML/HTML.
<p>A non-empty node is;
<ul>
<li>A node that isn't a text node. (probably an element node)
<li>A text node with non-whitespace content. [^ \f\n\r\t\u00A0\u2028\u2029].
<li>A text node with a Token.
</ul>
 
<p>When editing, what should you do about comment nodes???
<p>POST04: rename to be "nonWhiteSpaceNode"
<p>POST04: use ... if(node.isWhitespaceOnly) ... from DOM3

@return NodeFilter.FILTER_ACCEPT if not empty, .FILTER_REJECT if empty.
@type Integer
*/
__NodeFilter.nonEmptyText = function(node)
{
	var empty = NodeFilter.FILTER_REJECT;
	var notEmpty = NodeFilter.FILTER_ACCEPT;

	switch (node.nodeType) {
		case Node.COMMENT_NODE :
		case Node.CDATA_SECTION_NODE :
			return empty;
		case Node.ELEMENT_NODE :
			return notEmpty;
		case Node.TEXT_NODE :
			if(node.nodeValue.length == 0) // empty node is obviously empty!
				return empty;

			var re = eDOM.NON_WHITESPACE_RE;
			if( (re.test(node.nodeValue)) ) // any non white space visible characters 
				return notEmpty;

			re = eDOM.TOKEN_RE;
			if( (re.test(node.nodeValue)) ) // accept Token(s) 
				return notEmpty;

			return empty;

		default:
			return notEmpty;
	}
}

/****************** CONSTANTS *************************************************/
/**
eDOM local functions and constants.

To Do : need to sort out all references to Token so that it can be easily changed.
*/
function EDOM(){}
var eDOM = new EDOM();

EDOM.prototype.TOKEN = 160;
EDOM.prototype.NBSP = 160;
EDOM.prototype.SPACE = 32;
EDOM.prototype.LINEFEED = 10;
EDOM.prototype.ZWSPACE = 65279; // \uFEFF

/* Regular Expressions */
EDOM.prototype.TOKEN_RE = /\u00A0+/;       // 1 or more non-break space.
EDOM.prototype.NON_WHITESPACE_RE = /\S+/;  // 1 or more non whitespace character : [^ \f\n\r\t\u00A0\u2028\u2029].
/*
u00A0 = non-break space ; u2028 = line separator ; u2029 = paragraph separator
*/
 
EDOM.prototype.SPCHARS = "\f\n\r\t\u0020\u2028\u2029"; // NOTE: no nbsp - \u00A0 - consider rename to normalSpace

/**
Creates a Token text node.

@return Token
@type Text 
*/
EDOM.prototype.token = function() {
	return document.createTextNode(String.fromCharCode(eDOM.TOKEN));
}
/**
Creates a new line (linefeed 0Ah) text node;

@return newLine
@type Text 
*/
EDOM.prototype.newLine = function() {
	return document.createTextNode(String.fromCharCode(eDOM.LINEFEED));
}


//const SPCHARS = "\f\n\r\t\u0020\u2028\u2029"; // NOTE: no nbsp - \u00A0 - consider rename to normalSpace

const STRING_NBSP = "\u00A0";
const CHARCODE_NBSP = 160;
//const STRING_NBSP = "\u007B"; (test vals)
//const CHARCODE_NBSP = "123";
const CHARCODE_SPACE = 32;

const STRING_SPACE = "\u0020";
const CHARCODE_NEWLINE = "012";	
const STRING_LINEFEED = "\u000A";
const CHARCODE_TAB = "009";


/****************** Test Feedback *********************************************/

var feedback = false;
function __Feedbackack(text)
{
	if(!feedback)
		return;

	var feedbackArea = document.getElementById("feedbackArea");
	if(!feedbackArea)
		return;
	var feedback = document.createTextNode(text);
	var feedbackBlock = documentCreateXHTMLElement("div");
	feedbackBlock.appendChild(feedback);
	feedbackArea.appendChild(feedbackBlock);
}
