/* ***** BEGIN LICENSE BLOCK *****
 * Licensed under Version: MPL 1.1/GPL 2.0/LGPL 2.1
 * Full Terms at http://mozile.mozdev.org/license.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Playsophy code.
 *
 * The Initial Developer of the Original Code is Playsophy
 * Portions created by the Initial Developer are Copyright (C) 2002-2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *		Karl Guertin <grayrest@grayrest.com>
 *		James A. Overton <james@overton.ca>
 *		Christian Stocker (Bitflux)
 *		Paul Everitt
 *		David Palm
 *
 * ***** END LICENSE BLOCK ***** */

/**
@fileoverview
Mozile Undo Redo Javascript.

Created by James A. Overton
Created on 2004-03-24

<p>HISTORY:<br />
Some code in this file was first found in mozilehtmltb.js, while the 
majority was cut out of mozileTB.js and separated into different files.

<p>DESCRIPTION:<br />
This code handles the mozileUndo and mozileRedo functions, as well as the
mozileDo function.

<h2>Mozile Do, Undo and Redo</h2>

<p>Records the state of the current document into the undoStack for that document,
and increments the undoCounter (each document gets its own stack and counter).
All steps higher up in the stack are eliminated when a lower one is added.
If the undoMax number of steps is hit, the oldest entry is discarded.
The undoStack includes the name of the command and the whole BODY node.

<p>Yes, it is a brute force solution to copy the whole BODY, but it is the best
solution for multiple contentEditable areas. (Doing the whole HTML messes up
the scrollbar.) The performance seems fine, but will depend on the size of the
document and the size of the stack.
Of course this won't work for XML in general...

<p>To do: mozileDo is called before formatting commands.
If the command doesn't do anything should really remove last Do from the stack.
Could tidy up into an Undo object.

*/

const MOZILE_UNDO_NAME = 0;
const MOZILE_UNDO_NODE = 1;
const MOZILE_UNDO_FOCUS_NODE = 2;
const MOZILE_UNDO_FOCUS_OFFSET = 3;

/**
Create an undo object.
Initialisation required to add the undoStack etc. to a target window.
Note that keyCount is set to maximum so that the first keyboard entry will cause a save.

@param {integer} undoMax : maximum number of undo states to save.
@param {integer} keyMax : number of keypresses that will triger a "do". 
@constructor
*/
function MozileUndo( undoMax, keyMax ) {
	this.undoStack=new Array();
	this.undoCounter=0;
	this.undoMax=undoMax;
	this.DONE=0
	this.UNDONE=1;
	this.REDONE=2;
	this.undoState=this.DONE;
	this.keyMax=keyMax;
	this.keyCount=keyMax;
}

MozileUndo.prototype.NAME = 0;
MozileUndo.prototype.NODE = 1;
MozileUndo.prototype.FOCUS_NODE = 2;
MozileUndo.prototype.FOCUS_OFFSET = 3;

MozileUndo.prototype.__defineGetter__ (
	"justUndone",
	function () {return (this.undoState == this.UNDONE) ? true : false}
);
MozileUndo.prototype.__defineGetter__ (
	"justRedone",
	function () {return (this.undoState == this.REDONE) ? true : false}
);
MozileUndo.prototype.__defineGetter__ (
	"isUndo",
	function () {return (this.undoCounter > 0) ? true : false}
);
MozileUndo.prototype.__defineGetter__ (
	"isRedo",
	function () {return ( this.undoCounter < this.undoStack.length ) ? true : false}
);

/**
Do Typing.
Increment keyCount. 
Save to undo stack when keyCount reaches keyMax.
*/
function mozileDoTyping() {
	var undoObj = mozileTarget.mozileUndo;
	if( undoObj.keyCount >= undoObj.keyMax )
		mozileDo("Typing");
	undoObj.keyCount++;
}
/**
Save the current document state to the undo stack.
Reset the keyCount to zero.
"Do" should be called before an operation that changes the document.

@param {String} name
*/
function mozileDo(name) {
	var f = ["mozileUndoRedo.js","mozileDo"];
	mozileDebug(f,3,"Do: "+name);
	
	try{
		if ( !mozileTarget || !("mozileUndo" in mozileTarget) )
			return;
		//Prepare the data for the step to be added to the undo stack.
		var undoStep = new Array;
		undoStep[MOZILE_UNDO_NAME]=name;	//Store step name.
		
		//Get the last detected element node and index.
		var ename, enamespace, eoffset, elist;
		if ( (mozile.lastContextNode) && (mozile.lastContextNode.nodeType == Node.ELEMENT_NODE) ) {
			ename = mozile.lastContextNode.localName;
			enamespace = mozile.lastContextNode.namespaceURI;
			elist = mozileTarget.document.getElementsByTagNameNS( enamespace, ename );
			for ( var i=0; i<elist.length; i++ ) {
				if ( elist[i] === mozile.lastContextNode ) {
					eoffset = i;
					break;
				}
			}
		} else {
			mozileDebug(f,0,"No last context node");
		}
		//Store Element name and index.
		if ( (ename != undefined) && (eoffset != undefined) ) {
			mozileDebug(f,3,"namespace: "+enamespace+" /name: "+ename+" /offset: "+eoffset+ " /elist.length: "+elist.length);
			undoStep[MOZILE_UNDO_FOCUS_NODE] = ename;
			undoStep[MOZILE_UNDO_FOCUS_OFFSET] = eoffset;
		} else {
			mozileDebug(f,0,"No name/offset");
			undoStep[MOZILE_UNDO_FOCUS_NODE] = null;
			undoStep[MOZILE_UNDO_FOCUS_OFFSET] = null;
		}

		var getNode;
		//If x/html document clone the whole <body>...</body>
		if(mozileTarget.document.documentElement.tagName.toLowerCase()=="html") {
			mozileDebug(f,3,"X/HTML context: "+ mozileTarget.document.documentElement.tagName);
			getNode = mozileTarget.document.getElementsByTagName("body")[0];
		}
		else { //If not an x/html document clone the whole document element.
			mozileDebug(f,3,"XML context: "+ mozileTarget.document.documentElement.tagName);
			getNode=mozileTarget.document.documentElement;
		}
		undoStep[MOZILE_UNDO_NODE]=getNode.cloneNode(true);

		var undoObj = mozileTarget.mozileUndo;
		//Remove any unwanted "redo" entries.
		undoObj.undoStack=undoObj.undoStack.slice(0,undoObj.undoCounter);
		//If stack is full remove bottom entry to make 1 space at the top.
		if(undoObj.undoStack.length >= mozile.undoMax)
			undoObj.undoStack.shift();
		//Add new entry to stack.
		undoObj.undoStack.push(undoStep);
		//Update stack counter.
		undoObj.undoCounter=undoObj.undoStack.length;
		mozileUpdateUndoButtons();
		undoObj.keyCount=0;
		undoObj.undoState=undoObj.DONE;
		mozileTarget.document.MOZILE_CURRENT_CONFIG.saved = "false";
	} catch(e) {
		mozileDebug(f,1,"Error with the document:" + e);
	}
}

/**
Mozile Undo.
Moves down the undoStack from the curent undoCounter, replacing the current
BODY node with the one from the stack. If we are at the top of the stack
the Current state of the document is saved before the first undo.
*/
function mozileUndo() {
	var f = ["mozileUndoRedo.js","mozileUndo"];
	try {
		if ( !mozileTarget || !("mozileUndo" in mozileTarget) )
			return;
		var undoObj = mozileTarget.mozileUndo;
		if( !("undoCounter" in undoObj) || undoObj.undoCounter<1)
			return; //Nothing to undo.
  
		//If nothing on stack to "redo"
		var stackLength = undoObj.undoStack.length;
		if(undoObj.undoCounter==stackLength) {
			if ( (undoObj.undoStack[stackLength-1][MOZILE_UNDO_NAME] != "Current") || (undoObj.keyCount > 0) ) {
				//Save the current state, but only once!
				mozileDo('Current');
				//No need to undo current state.
				undoObj.undoCounter--;
			}
		}
		var c = undoObj.undoCounter-1;
		if ( undoObj.justRedone )
			c--;

		var undoStep=undoObj.undoStack[c];
		mozileReplaceNode(undoStep[MOZILE_UNDO_NODE]);
		mozileRestoreFocus( undoStep );

		undoObj.undoCounter=c;
		undoObj.undoState=undoObj.UNDONE;
		mozileUpdateUndoButtons();
	} catch (e) {
		mozileDebug(f,1,"Exception: "+e);
	}
}

/**
Mozile Redo.
Moves up the undoStack from the current undoCounter,
replacing the current BODY node with the one in the stack.
*/
function mozileRedo() {
	var f = ["mozileUndoRedo.js","mozileRedo"];
	try {
		if ( !mozileTarget || !("mozileUndo" in mozileTarget) )
			return;
		var undoObj = mozileTarget.mozileUndo;
		var c = undoObj.undoCounter;
		var s = undoObj.undoStack.length;
		if( c < s ) {
			//If just undone, need to skip an entry on the stack.
			if ( undoObj.justUndone )
				c++;
			var undoStep=undoObj.undoStack[c];
			//Replace document content
			mozileReplaceNode(undoStep[MOZILE_UNDO_NODE]);
			mozileRestoreFocus( undoStep );
			//Update stack counter.
			undoObj.undoCounter = c+1;
			undoObj.undoState=undoObj.REDONE;
			mozileUpdateUndoButtons();
		}
	} catch (e) {
		mozileDebug(f,1,"Exception: "+e);
	}
}

/**
Finds the element in the document and restores the focus to the end of the element.
*/	
function mozileRestoreFocus( undoStep ) {
	var f = ["mozileUndoRedo.js","mozileRestoreFocus"];
	try {
		var fnode = undoStep[MOZILE_UNDO_FOCUS_NODE];
		var foffset = undoStep[MOZILE_UNDO_FOCUS_OFFSET];
		if ( (fnode != null) && (foffset != null) ) {
			var elist = mozileTarget.document.getElementsByTagName( fnode );
			if ( foffset < elist.length ) {
				var currentSelection = mozileTarget.getSelection();
				currentSelection.removeAllRanges();
				var range = mozileTarget.document.createRange();
				var e = elist[foffset];
				var top = e.userModifiableContext;
				if (top) { 
					var ip = e.lastInsertionPoint( top );
					if (ip) {
						range.selectInsertionPoint( ip );
						currentSelection.addRange( range );
					}
				}
			}
		} else {
			mozileDebug(f,1,"No focus to restore!");
		}
	} catch (e) {
		mozileDebug(f,1,"Exception: "+e);
	}
}	

/**
Replace a special target node with the replacement.
The target will depend on the content, for HTML it is the body tag,
for XML it is currently the document element.
*/
function mozileReplaceNode(replacement) {
	var f = ["mozileUndoRedo.js","mozileReplaceNode"];	
	mozileDebug(f,3,"Replacing Node "+ replacement);
	var target, rChildren, rLength, replacementFragment, i;
	
	if(mozileTarget.document.documentElement.tagName.toLowerCase()=="html") {
		mozileDebug(f,3,"X/HTML context");
		try {
			replacementFragment = mozileTarget.document.createDocumentFragment();

			rChildren = replacement.childNodes;
			rLength = rChildren.length;
			for ( i=0; i<rLength; i++ ) {
				replacementFragment.appendChild( rChildren[i].cloneNode(true) );
			}

			target = mozileTarget.document.getElementsByTagName("body")[0];
			while ( target.hasChildNodes() ) { //remove all child nodes from body.
				target.removeChild( target.firstChild );
			}
			target.appendChild( replacementFragment );

		} catch (e) {
			mozileDebug(f,0,"Exception! "+e);
		}
	}
	else {
		mozileDebug(f,3,"XML context");
		//Note: cloning and replacing the whole documentElement locks the browser scrollbars.
		//Using a document fragment seems to cure this.
		try {
			replacementFragment = mozileTarget.document.createDocumentFragment();

			rChildren = replacement.childNodes;
			rLength = rChildren.length;
			for ( i=0; i<rLength; i++ ) {
				replacementFragment.appendChild( rChildren[i].cloneNode(true) );
			}

			target = mozileTarget.document.documentElement;
			while ( target.hasChildNodes() ) { //remove all child nodes from body.
				target.removeChild( target.firstChild );
			}
			target.appendChild( replacementFragment );

		} catch (e) {
			mozileDebug(f,1,"Exception! "+e);
		}

	}
}


